"""Tests for checking the functionality of typed components, model output thunks and sampling results."""

from typing import get_args

import pytest

import mellea.stdlib.functional as mfuncs
from mellea import MelleaSession, start_session
from mellea.backends.model_ids import IBM_GRANITE_4_HYBRID_MICRO
from mellea.backends.ollama import OllamaModelBackend
from mellea.core import (
    CBlock,
    Component,
    ComponentParseError,
    Context,
    ModelOutputThunk,
    Requirement,
    ValidationResult,
)
from mellea.stdlib.components import Instruction, Message
from mellea.stdlib.context import ChatContext, SimpleContext
from mellea.stdlib.sampling import BaseSamplingStrategy


class FloatComp(Component[float]):
    def __init__(self, value: str) -> None:
        self.value = value

    def parts(self) -> list[Component | CBlock]:
        return []

    def format_for_llm(self) -> str:
        return self.value

    def _parse(self, computed: ModelOutputThunk) -> float:
        if computed.value is None:
            return -1
        return float(computed.value)


class IntComp(FloatComp, Component[int]):
    def _parse(self, computed: ModelOutputThunk) -> int:
        if computed.value is None:
            return -1
        try:
            return int(computed.value)
        except Exception:
            return -2


class ExceptionRaisingComp(Component[int]):
    def parts(self) -> list[Component | CBlock]:
        return []

    def format_for_llm(self) -> str:
        return ""

    def _parse(self, computed: ModelOutputThunk) -> int:
        raise ValueError("random error")


@pytest.fixture(scope="module")
def backend(gh_run: int):
    """Shared backend."""
    if gh_run == 1:
        return OllamaModelBackend(
            model_id=IBM_GRANITE_4_HYBRID_MICRO.ollama_name  # type: ignore
        )
    else:
        return OllamaModelBackend(model_id=IBM_GRANITE_4_HYBRID_MICRO.ollama_name)  # type: ignore


@pytest.fixture(scope="module")
def session(backend) -> MelleaSession:
    return MelleaSession(backend=backend, ctx=SimpleContext())


def test_mot_init_typing():
    mot = ModelOutputThunk[float](value="1")
    assert hasattr(mot, "__orig_class__"), (
        "mots are generics and should have this field"
    )
    assert get_args(mot.__orig_class__)[0] is float, (  # type: ignore
        f"expected float, got {get_args(mot.__orig_class__)[0]} as mot type"  # type: ignore
    )  # type: ignore

    unknown_mot = ModelOutputThunk(value="2")
    assert not hasattr(unknown_mot, "__orig_class__"), (
        "unknown mots / mots with no type defined at instantiate don't have this attribute"
    )


def test_simple_component_parsing():
    fc = FloatComp(value="generate a float")
    mot = ModelOutputThunk[float](value="1")
    assert fc.parse(mot) == 1
    assert isinstance(fc.parse(mot), float)


def test_subclassed_component_parsing():
    ic = IntComp("generate an int")
    mot = ModelOutputThunk[float](value="1")
    assert ic.parse(mot) == 1


def test_component_parsing_fails():
    erc = ExceptionRaisingComp()
    mot = ModelOutputThunk[float](value="1")

    with pytest.raises(ComponentParseError):
        _ = erc.parse(mot) == 1


def test_incorrect_type_override():
    with pytest.raises(TypeError):
        Instruction[int](description="this is an instruction")  # type: ignore


# Marking as qualitative for now since there's so much generation required for this.
# Uses granite4:micro-h (3B hybrid, lightweight) in local mode
@pytest.mark.qualitative
@pytest.mark.ollama
@pytest.mark.requires_gpu
@pytest.mark.llm
async def test_generating(session):
    m = session
    ic = IntComp("generate an int")

    out, _ = mfuncs.act(ic, context=ChatContext(), backend=m.backend, strategy=None)
    assert isinstance(out.parsed_repr, int)

    # `out` typed as ModelOutputThunk[str]
    out, _ = await m.backend.generate_from_context(
        CBlock("Say Hello!"), ctx=ChatContext()
    )
    await out.avalue()
    assert isinstance(out.parsed_repr, str)

    # `out` typed as ModelOutputThunk[float]
    out, _ = await m.backend.generate_from_context(ic, ctx=ChatContext())
    await out.avalue()
    assert isinstance(out.parsed_repr, int)

    # `out` typed as ModelOutputThunk[float | str]
    out = await m.backend.generate_from_raw([ic, CBlock("")], ctx=ChatContext())
    for result in out:
        await result.avalue()
    assert isinstance(out[0].parsed_repr, int)
    assert isinstance(out[1].parsed_repr, str)

    # `out` typed as ModelOutputThunk[float]
    out = await m.backend.generate_from_raw([ic, ic], ctx=ChatContext())
    for result in out:
        await result.avalue()
        assert isinstance(result.parsed_repr, int)

    # `out` typed as ModelOutputThunk[str]
    out = await m.backend.generate_from_raw([CBlock("")], ctx=ChatContext())
    for result in out:
        await result.avalue()
        assert isinstance(result.parsed_repr, str)


@pytest.mark.qualitative
@pytest.mark.ollama
@pytest.mark.requires_gpu
@pytest.mark.requires_heavy_ram
@pytest.mark.llm
def test_message_typing(session):
    m = session
    user_message = Message("user", "Hello!")
    response = m.act(user_message)
    assert response.parsed_repr is not None
    assert isinstance(response.parsed_repr, Message)

    second_response = m.act(response.parsed_repr)
    assert second_response.parsed_repr is not None
    assert isinstance(second_response.parsed_repr, Message)


@pytest.mark.qualitative
@pytest.mark.ollama
@pytest.mark.requires_gpu
@pytest.mark.requires_heavy_ram
@pytest.mark.llm
async def test_generating_with_sampling(session):
    m = session
    m = start_session()

    class CustomSamplingStrat(BaseSamplingStrategy):
        @staticmethod
        def select_from_failure(
            sampled_actions: list[Component],
            sampled_results: list[ModelOutputThunk],
            sampled_val: list[list[tuple[Requirement, ValidationResult]]],
        ) -> int:
            return len(sampled_actions) - 1

        @staticmethod
        def repair(
            old_ctx: Context,
            new_ctx: Context,
            past_actions: list[Component],
            past_results: list[ModelOutputThunk],
            past_val: list[list[tuple[Requirement, ValidationResult]]],
        ) -> tuple[Component, Context]:
            return Instruction("print another number 100 greater"), old_ctx

    css = CustomSamplingStrat(loop_budget=3)
    out = await css.sample(
        action=IntComp("2000"),
        context=ChatContext(),
        backend=m.backend,
        requirements=[
            Requirement(
                None, validation_fn=lambda x: ValidationResult(False), check_only=True
            )
        ],
    )

    # Even though the intermediate actions are Instructions, the parsed_reprs at each stage
    # are ints.
    for result in out.sample_generations:
        assert isinstance(result.parsed_repr, int), (
            "model output thunks should have the correct parsed_repr type"
        )

    for action in out.sample_actions[1:]:
        assert isinstance(action, Instruction), (
            "repair strategy should force repaired actions to be Instructions"
        )


if __name__ == "__main__":
    pytest.main([__file__])



"""Module for working with Requirements."""

# Import from core for ergonomics.
from ...core import Requirement, ValidationResult, default_output_to_bool
from .md import as_markdown_list, is_markdown_list, is_markdown_table
from .python_reqs import PythonExecutionReq
from .requirement import (
    ALoraRequirement,
    LLMaJRequirement,
    check,
    req,
    reqify,
    requirement_check_to_bool,
    simple_validate,
)
from .tool_reqs import tool_arg_validator, uses_tool

__all__ = [
    "ALoraRequirement",
    "LLMaJRequirement",
    "PythonExecutionReq",
    "Requirement",
    "ValidationResult",
    "as_markdown_list",
    "check",
    "default_output_to_bool",
    "is_markdown_list",
    "is_markdown_table",
    "req",
    "reqify",
    "requirement_check_to_bool",
    "simple_validate",
    "tool_arg_validator",
    "uses_tool",
]



# pytest: ollama, llm

from docs.examples.helper import w
from mellea import start_session
from mellea.backends import ModelOption

# create a session using Granite 4 Micro 3B on Ollama and a simple context [see below]
m = start_session(model_options={ModelOption.MAX_NEW_TOKENS: 200})

# write an email with automatic requirement checking.
email_v1 = m.instruct(
    "Write an email to invite all interns to the office party.",
    requirements=["be formal", "Use 'Dear Interns' as greeting."],
)

# print result
print(f"***** email ****\n{w(email_v1)}\n*******")

# # optionally, for debugging.. print last prompt
# print(f"Prompt:\n{w(m.last_prompt())}")



from typing import TypedDict


class ICLExample(TypedDict):
    execution_plan: list[str]
    available_content_variables: list[str]
    target_subtask: str
    subtask_prompt_template: str


class ICLExampleGroup(TypedDict):
    task_prompt: str
    examples_items: list[ICLExample]



"""Unit tests for SOFAISamplingStrategy."""

from unittest.mock import MagicMock

import pytest

from mellea.backends import Backend
from mellea.core import Requirement, ValidationResult
from mellea.stdlib.components import Instruction, Message, ModelOutputThunk
from mellea.stdlib.context import ChatContext
from mellea.stdlib.sampling import SOFAISamplingStrategy


class TestSOFAIInit:
    """Test SOFAISamplingStrategy initialization."""

    def test_init_valid_backends(self):
        """Test initialization with valid backends."""
        s1 = MagicMock(spec=Backend)
        s2 = MagicMock(spec=Backend)

        strategy = SOFAISamplingStrategy(s1_solver_backend=s1, s2_solver_backend=s2)

        assert strategy.s1_solver_backend is s1
        assert strategy.s2_solver_backend is s2
        assert strategy.loop_budget == 3  # default
        assert strategy.s2_solver_mode == "fresh_start"  # default

    def test_init_with_judge_backend(self):
        """Test initialization with judge backend."""
        s1 = MagicMock(spec=Backend)
        s2 = MagicMock(spec=Backend)
        judge = MagicMock(spec=Backend)

        strategy = SOFAISamplingStrategy(
            s1_solver_backend=s1,
            s2_solver_backend=s2,
            judge_backend=judge,
            feedback_strategy="all_errors",
        )

        assert strategy.judge_backend is judge
        assert strategy.feedback_strategy == "all_errors"

    def test_init_invalid_s1_backend_raises(self):
        """Test that non-Backend s1_solver raises TypeError."""
        s2 = MagicMock(spec=Backend)

        with pytest.raises(TypeError):
            SOFAISamplingStrategy(
                s1_solver_backend="not a backend",  # type: ignore
                s2_solver_backend=s2,
            )

    def test_init_invalid_s2_backend_raises(self):
        """Test that non-Backend s2_solver raises TypeError."""
        s1 = MagicMock(spec=Backend)

        with pytest.raises(TypeError):
            SOFAISamplingStrategy(
                s1_solver_backend=s1,
                s2_solver_backend="not a backend",  # type: ignore
            )

    def test_init_invalid_judge_backend_raises(self):
        """Test that non-Backend judge raises TypeError."""
        s1 = MagicMock(spec=Backend)
        s2 = MagicMock(spec=Backend)

        with pytest.raises(TypeError):
            SOFAISamplingStrategy(
                s1_solver_backend=s1,
                s2_solver_backend=s2,
                judge_backend="not a backend",  # type: ignore
            )

    def test_init_invalid_loop_budget_raises(self):
        """Test that loop_budget <= 0 raises ValueError."""
        s1 = MagicMock(spec=Backend)
        s2 = MagicMock(spec=Backend)

        with pytest.raises(ValueError):
            SOFAISamplingStrategy(
                s1_solver_backend=s1, s2_solver_backend=s2, loop_budget=0
            )

    def test_init_all_s2_modes(self):
        """Test all three s2_solver_mode options."""
        s1 = MagicMock(spec=Backend)
        s2 = MagicMock(spec=Backend)

        for mode in ["fresh_start", "continue_chat", "best_attempt"]:
            strategy = SOFAISamplingStrategy(
                s1_solver_backend=s1,
                s2_solver_backend=s2,
                s2_solver_mode=mode,  # type: ignore
            )
            assert strategy.s2_solver_mode == mode


class TestSOFAIRepair:
    """Test SOFAISamplingStrategy repair method."""

    def test_repair_creates_message_with_feedback(self):
        """Test that repair creates a Message with feedback from ValidationResult."""
        old_ctx = MagicMock(spec=ChatContext)
        new_ctx = MagicMock(spec=ChatContext)

        past_actions = [MagicMock(spec=Instruction)]
        past_results = [MagicMock(spec=ModelOutputThunk)]

        # Create validation results with reasons
        req1 = Requirement(description="Be formal")
        val1 = ValidationResult(False, reason="Output was too casual")
        req2 = Requirement(description="Use greeting")
        val2 = ValidationResult(True)  # Passed

        past_val = [[(req1, val1), (req2, val2)]]

        next_action, returned_ctx = SOFAISamplingStrategy.repair(
            old_ctx, new_ctx, past_actions, past_results, past_val
        )

        assert isinstance(next_action, Message)
        assert next_action.role == "user"
        assert "Output was too casual" in next_action.content
        assert "Be formal" not in next_action.content  # Uses reason, not description
        assert returned_ctx is new_ctx

    def test_repair_uses_description_when_no_reason(self):
        """Test that repair uses description when ValidationResult has no reason."""
        old_ctx = MagicMock(spec=ChatContext)
        new_ctx = MagicMock(spec=ChatContext)

        past_actions = [MagicMock(spec=Instruction)]
        past_results = [MagicMock(spec=ModelOutputThunk)]

        req1 = Requirement(description="Be formal")
        val1 = ValidationResult(False)  # No reason

        past_val = [[(req1, val1)]]

        next_action, _ = SOFAISamplingStrategy.repair(
            old_ctx, new_ctx, past_actions, past_results, past_val
        )

        assert isinstance(next_action, Message)
        assert "Be formal" in next_action.content  # Falls back to description


class TestSOFAISelectFromFailure:
    """Test SOFAISamplingStrategy select_from_failure method."""

    def test_select_from_failure_returns_last(self):
        """Test that select_from_failure returns last index (-1)."""
        actions = [MagicMock(), MagicMock(), MagicMock()]
        results = [MagicMock(), MagicMock(), MagicMock()]
        validations = [[], [], []]

        index = SOFAISamplingStrategy.select_from_failure(actions, results, validations)
        assert index == -1


class TestSOFAISelectBestAttempt:
    """Test SOFAISamplingStrategy _select_best_attempt method."""

    def test_select_best_attempt_single(self):
        """Test with single attempt."""
        req1 = Requirement(description="R1")
        val1 = ValidationResult(True)

        sampled_val = [[(req1, val1)]]
        index = SOFAISamplingStrategy._select_best_attempt(sampled_val)
        assert index == 0

    def test_select_best_attempt_prefers_more_passes(self):
        """Test that attempt with more passes is selected."""
        req1 = Requirement(description="R1")
        req2 = Requirement(description="R2")

        # First attempt: 1 pass
        val1_1 = ValidationResult(True)
        val1_2 = ValidationResult(False)

        # Second attempt: 2 passes
        val2_1 = ValidationResult(True)
        val2_2 = ValidationResult(True)

        sampled_val = [
            [(req1, val1_1), (req2, val1_2)],
            [(req1, val2_1), (req2, val2_2)],
        ]

        index = SOFAISamplingStrategy._select_best_attempt(sampled_val)
        assert index == 1

    def test_select_best_attempt_prefers_later_on_tie(self):
        """Test that later attempt is selected on tie."""
        req1 = Requirement(description="R1")

        # Both attempts: 1 pass
        val1 = ValidationResult(True)
        val2 = ValidationResult(True)

        sampled_val = [[(req1, val1)], [(req1, val2)]]

        index = SOFAISamplingStrategy._select_best_attempt(sampled_val)
        assert index == 1  # Prefers later


class TestSOFAIParseJudgment:
    """Test SOFAISamplingStrategy _parse_judgment method."""

    def test_parse_yes(self):
        """Test parsing 'Yes' responses."""
        assert SOFAISamplingStrategy._parse_judgment("Yes") is True
        assert SOFAISamplingStrategy._parse_judgment("yes") is True
        assert SOFAISamplingStrategy._parse_judgment("YES") is True
        assert (
            SOFAISamplingStrategy._parse_judgment("Yes, the requirement is satisfied.")
            is True
        )

    def test_parse_no(self):
        """Test parsing 'No' responses."""
        assert SOFAISamplingStrategy._parse_judgment("No") is False
        assert SOFAISamplingStrategy._parse_judgment("no") is False
        assert SOFAISamplingStrategy._parse_judgment("NO") is False
        assert SOFAISamplingStrategy._parse_judgment("No, there are errors.") is False

    def test_parse_yes_in_first_line(self):
        """Test that 'yes' in first line returns True."""
        assert SOFAISamplingStrategy._parse_judgment("The answer is yes.") is True

    def test_parse_no_first_line_default(self):
        """Test that ambiguous response defaults to False."""
        assert SOFAISamplingStrategy._parse_judgment("Maybe") is False
        assert SOFAISamplingStrategy._parse_judgment("I'm not sure") is False


class TestSOFAIExtractFeedback:
    """Test SOFAISamplingStrategy _extract_feedback method."""

    def test_extract_from_tags(self):
        """Test extracting feedback from <feedback> tags."""
        response = "No\n<feedback>The error is X.</feedback>"
        assert SOFAISamplingStrategy._extract_feedback(response) == "The error is X."

    def test_extract_multiline(self):
        """Test extracting multiline feedback."""
        response = """No
<feedback>
Error 1: X
Error 2: Y
</feedback>"""
        feedback = SOFAISamplingStrategy._extract_feedback(response)
        assert "Error 1: X" in feedback
        assert "Error 2: Y" in feedback

    def test_fallback_without_tags(self):
        """Test fallback to full response when no tags."""
        response = "No, the output is wrong."
        assert (
            SOFAISamplingStrategy._extract_feedback(response)
            == "No, the output is wrong."
        )


@pytest.mark.qualitative
@pytest.mark.ollama
@pytest.mark.llm
class TestSOFAIIntegration:
    """Integration tests for SOFAISamplingStrategy.

    These tests require actual LLM backends and are marked as qualitative.
    Uses llama3.2:1b (lightweight, no heavy RAM needed).
    """

    def test_sofai_with_ollama(self, gh_run):
        """Test SOFAI with Ollama backends."""
        from mellea import start_session
        from mellea.backends.ollama import OllamaModelBackend

        # Use smaller models for testing
        s1 = OllamaModelBackend(model_id="llama3.2:1b")
        s2 = OllamaModelBackend(model_id="llama3.2:1b")

        strategy = SOFAISamplingStrategy(
            s1_solver_backend=s1, s2_solver_backend=s2, loop_budget=2
        )

        m = start_session("ollama", model_id="llama3.2:1b", ctx=ChatContext())

        result = m.instruct(
            "Write a haiku about coding.",
            requirements=["The haiku should have 3 lines"],
            strategy=strategy,
            return_sampling_results=True,
        )

        assert result is not None
        assert hasattr(result, "success")
        assert len(result.sample_generations) >= 1


if __name__ == "__main__":
    pytest.main(["-v", __file__])



import json
from unittest.mock import MagicMock, patch

import pytest
import requests

from mellea.helpers import is_vllm_server_with_structured_output

BASE_URL = "http://localhost:8000/v1"
HEADERS = {"Authorization": "Bearer test-key"}


def _mock_version_response(version_string, status_code=200):
    mock_resp = MagicMock(spec=requests.Response)
    mock_resp.ok = 200 <= status_code < 400
    mock_resp.status_code = status_code
    mock_resp.text = json.dumps({"version": version_string})
    return mock_resp


@patch("mellea.helpers.server_type.requests.get")
def test_version_above_threshold(mock_get):
    mock_get.return_value = _mock_version_response("0.16.0")
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is True
    mock_get.assert_called_once_with("http://localhost:8000/version", headers=HEADERS)


@patch("mellea.helpers.server_type.requests.get")
def test_version_equal_to_threshold(mock_get):
    mock_get.return_value = _mock_version_response("0.12.0")
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is True


@patch("mellea.helpers.server_type.requests.get")
def test_version_below_threshold(mock_get):
    mock_get.return_value = _mock_version_response("0.11.9")
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_version_much_lower(mock_get):
    mock_get.return_value = _mock_version_response("0.6.0")
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_release_candidate_version_above(mock_get):
    mock_get.return_value = _mock_version_response("0.16.0rc1.dev172+gddf69")
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is True


@patch("mellea.helpers.server_type.requests.get")
def test_http_error_response(mock_get):
    mock_resp = MagicMock(spec=requests.Response)
    mock_resp.ok = False
    mock_resp.status_code = 500
    mock_get.return_value = mock_resp
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_http_404_response(mock_get):
    mock_resp = MagicMock(spec=requests.Response)
    mock_resp.ok = False
    mock_resp.status_code = 404
    mock_get.return_value = mock_resp
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_no_version_key_in_response(mock_get):
    mock_resp = MagicMock(spec=requests.Response)
    mock_resp.ok = True
    mock_resp.text = json.dumps({"status": "ok"})
    mock_get.return_value = mock_resp
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_empty_json_response(mock_get):
    mock_resp = MagicMock(spec=requests.Response)
    mock_resp.ok = True
    mock_resp.text = json.dumps({})
    mock_get.return_value = mock_resp
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_invalid_json_response(mock_get):
    mock_resp = MagicMock(spec=requests.Response)
    mock_resp.ok = True
    mock_resp.text = "not json at all"
    mock_get.return_value = mock_resp
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_invalid_version_string(mock_get):
    mock_resp = MagicMock(spec=requests.Response)
    mock_resp.ok = True
    mock_resp.text = json.dumps({"version": "not-a-version"})
    mock_get.return_value = mock_resp
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_connection_error(mock_get):
    mock_get.side_effect = requests.ConnectionError("Connection refused")
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_timeout_error(mock_get):
    mock_get.side_effect = requests.Timeout("Request timed out")
    assert is_vllm_server_with_structured_output(BASE_URL, HEADERS) is False


@patch("mellea.helpers.server_type.requests.get")
def test_url_strips_v1_suffix(mock_get):
    mock_get.return_value = _mock_version_response("0.14.0")
    is_vllm_server_with_structured_output("http://myserver:9000/v1", HEADERS)
    mock_get.assert_called_once_with("http://myserver:9000/version", headers=HEADERS)


@patch("mellea.helpers.server_type.requests.get")
def test_url_strips_v1_slash_suffix(mock_get):
    mock_get.return_value = _mock_version_response("0.14.0")
    is_vllm_server_with_structured_output("http://myserver:9000/v1/", HEADERS)
    mock_get.assert_called_once_with("http://myserver:9000/version", headers=HEADERS)


@patch("mellea.helpers.server_type.requests.get")
def test_url_without_v1(mock_get):
    mock_get.return_value = _mock_version_response("0.14.0")
    is_vllm_server_with_structured_output("http://myserver:9000", HEADERS)
    mock_get.assert_called_once_with("http://myserver:9000/version", headers=HEADERS)


if __name__ == "__main__":
    import pytest

    pytest.main([__file__])



"""Testing functions for budget forcing generation."""

import pytest

from mellea import MelleaSession, start_session
from mellea.backends import ModelOption
from mellea.backends.model_ids import OPENAI_GPT_OSS_20B
from mellea.core import CBlock
from mellea.stdlib.sampling.budget_forcing import BudgetForcingSamplingStrategy

MODEL_ID = OPENAI_GPT_OSS_20B

# Module-level markers: gpt-oss:20b is a 20B model requiring heavy resources
pytestmark = [
    pytest.mark.ollama,
    pytest.mark.requires_gpu,
    pytest.mark.requires_heavy_ram,
    pytest.mark.llm,
    pytest.mark.qualitative,
]


@pytest.fixture(scope="module")
def m_session(gh_run):
    """Start default Mellea's session."""
    if gh_run == 1:  # on github
        m = start_session(
            "ollama", model_id=MODEL_ID, model_options={ModelOption.MAX_NEW_TOKENS: 5}
        )
    else:
        m = start_session("ollama", model_id=MODEL_ID)
    yield m
    del m


def test_think_big(m_session: MelleaSession, gh_run: int):
    """Tests big thinking budget."""
    # if on github we can run big thinking mode
    if gh_run == 1:
        pytest.skip("Skipping big_thinking runs in gh workflows.")

    prompt = "What is the smallest positive integer $n$ such that all the roots of $z^4 + z^2 + 1 = 0$ are $n^{\\text{th}}$ roots of unity?"
    prompt_suffix = "\nPlease reason step by step, use \n\n to end each step, and put your final answer within \\boxed{}."
    action = CBlock(value=prompt + prompt_suffix)
    THINK_MAX_TOKENS = 2048
    ANSWER_MAX_TOKENS = 512

    strategy = BudgetForcingSamplingStrategy(
        think_max_tokens=THINK_MAX_TOKENS,
        answer_max_tokens=ANSWER_MAX_TOKENS,
        start_think_token="<think>",
        end_think_token="</think>",
        think_more_suffix="\nWait, let's think more carefully",
        answer_suffix="The final answer is:",
        requirements=None,
    )
    result = m_session.instruct(action, strategy=strategy)  # type: ignore

    print("\n******\nThink big:")
    print(str(result))


def test_think_little(m_session: MelleaSession, gh_run: int):
    """Tests little thinking budget."""
    prompt = "Compute 1+1?"
    prompt_suffix = "\nPlease reason step by step, use \n\n to end each step, and put your final answer within \\boxed{}."
    action = CBlock(value=prompt + prompt_suffix)
    THINK_MAX_TOKENS = 16
    ANSWER_MAX_TOKENS = 8
    if gh_run == 1:  # on github
        THINK_MAX_TOKENS = 0
        ANSWER_MAX_TOKENS = 5

    strategy = BudgetForcingSamplingStrategy(
        think_max_tokens=THINK_MAX_TOKENS,
        answer_max_tokens=ANSWER_MAX_TOKENS,
        start_think_token="<think>",
        end_think_token="</think>",
        answer_suffix="The final answer is: \\boxed{",
        requirements=None,
    )
    result = m_session.instruct(action, strategy=strategy)  # type: ignore

    print("\n******\nThink little:")
    print(str(result))


if __name__ == "__main__":
    pytest.main(["-s", __file__])



import pytest

from mellea.stdlib.components import Message, as_chat_history
from mellea.stdlib.context import ChatContext
from mellea.stdlib.session import start_session

# Mark all tests as requiring Ollama (start_session defaults to Ollama)
pytestmark = [pytest.mark.ollama, pytest.mark.llm]


@pytest.fixture(scope="function")
def linear_session():
    """Session with linear context for chat tests."""
    session = start_session(ctx=ChatContext())
    yield session
    session.reset()


@pytest.fixture(scope="function")
def simple_session():
    """Session with simple context for chat tests."""
    session = start_session()
    yield session
    session.reset()


def test_chat_view_linear_ctx(linear_session):
    linear_session.chat("What is 1+1?")
    linear_session.chat("What is 2+2?")
    assert len(as_chat_history(linear_session.ctx)) == 4
    assert all(isinstance(x, Message) for x in as_chat_history(linear_session.ctx))
    assert len(linear_session.ctx.view_for_generation()) == 4


# @pytest.mark.skip("linearize() returns [] for a SimpleContext... that's going to be annoying.")
def test_chat_view_simple_ctx(simple_session):
    simple_session.chat("What is 1+1?")
    simple_session.chat("What is 2+2?")
    assert len(as_chat_history(simple_session.ctx)) == 4
    assert all(isinstance(x, Message) for x in as_chat_history(simple_session.ctx))
    assert len(simple_session.ctx.view_for_generation()) == 0


if __name__ == "__main__":
    pytest.main([__file__])



# pytest: ollama, qualitative, llm, requires_heavy_ram

from mellea.backends import model_ids
from mellea.backends.model_ids import IBM_GRANITE_4_HYBRID_MICRO
from mellea.stdlib.components.docs.richdocument import RichDocument

rd = RichDocument.from_document_file("https://arxiv.org/pdf/1906.04043")

from mellea.stdlib.components.docs.richdocument import Table

table1: Table = rd.get_tables()[0]
print(table1.to_markdown())

from mellea import start_session
from mellea.backends import ModelOption

m = start_session(model_id=model_ids.META_LLAMA_3_2_3B)
for seed in [x * 12 for x in range(5)]:
    table2 = m.transform(
        table1,
        "Add a column 'Model' that extracts which model was used in Feature description or 'None' if none.",
        model_options={ModelOption.SEED: seed},
    )
    if isinstance(table2, Table):
        print(table2.to_markdown())
        break
    else:
        print("==== TRYING AGAIN after non-useful output.====")



import pytest

from mellea.backends import ModelOption
from mellea.core import ModelOutputThunk
from mellea.stdlib.components import Message
from mellea.stdlib.functional import aact, ainstruct, avalidate, instruct
from mellea.stdlib.requirements import req
from mellea.stdlib.session import start_session

pytestmark = [pytest.mark.ollama, pytest.mark.llm]


@pytest.fixture(scope="module")
def m_session(gh_run):
    m = start_session(model_options={ModelOption.MAX_NEW_TOKENS: 5})
    yield m
    del m


def test_func_context(m_session):
    initial_ctx = m_session.ctx
    backend = m_session.backend

    out, ctx = instruct("Write a sentence.", initial_ctx, backend)
    assert initial_ctx is not ctx
    assert ctx._data is out


async def test_aact(m_session):
    initial_ctx = m_session.ctx
    backend = m_session.backend

    out, ctx = await aact(Message(role="user", content="hello"), initial_ctx, backend)

    assert initial_ctx is not ctx
    assert ctx._data is out


async def test_ainstruct(m_session):
    initial_ctx = m_session.ctx
    backend = m_session.backend

    out, ctx = await ainstruct("Write a sentence", initial_ctx, backend)

    assert initial_ctx is not ctx
    assert ctx._data is out


async def test_avalidate(m_session):
    initial_ctx = m_session.ctx
    backend = m_session.backend

    val_result = await avalidate(
        reqs=[req("Be formal."), req("Avoid telling jokes.")],
        context=initial_ctx,
        backend=backend,
        output=ModelOutputThunk("Here is an output."),
    )

    assert len(val_result) == 2
    assert val_result[0] is not None


if __name__ == "__main__":
    pytest.main([__file__])



# pytest: ollama, llm

from mellea.stdlib.components.docs.richdocument import TableQuery
from mellea.stdlib.components.mify import MifiedProtocol, mify
from mellea.stdlib.session import start_session


# Mify works on python objects and classes. Apply it to your own
# custom class or object to start working with mellea.
@mify
class MyCustomerClass:
    def __init__(self, name: str, last_purchase: str) -> None:
        self.name = name
        self.last_purchase = last_purchase


# Now when you instantiate an object of that class, it will also
# have the fields and members necessary for working with mellea.
c = MyCustomerClass("Jack", "Beans")
assert isinstance(c, MifiedProtocol)


# You can also mify objects ad hoc.
class MyStoreClass:
    def __init__(self, purchases: list[str]) -> None:
        self.purchases: list[str]


store = MyStoreClass(["Beans", "Soil", "Watering Can"])
mify(store)
assert isinstance(store, MifiedProtocol)

# Now, you can use these objects in MelleaSessions.
store.format_for_llm()
m = start_session()
m.act(store)


# However, unless your object/class has a __str__ function,
# this won't do much good by itself. You need to specify how
# mellea should process these objects as text. You can do this by
# parameterizing mify.
@mify(stringify_func=lambda x: f"Chain Location: {x.location}")  # type: ignore
class MyChain:
    def __init__(self, location: str):
        self.location = location


# M operations will now utilize that string representation of the
# object when interacting with it.
m.query(MyChain("Northeast"), "Where is my chain located?")


# For more complicated representations, you can utilize mify
# to interact with our templating system. Here, we know that a
# TableQuery calls its underlying object's to_markdown function.
# Since our class has the same process, we can use that template.
# We can also specify that our class should use either a template with it's own
# class name or the Table template when not querying.
@mify(query_type=TableQuery, template_order=["*", "Table"])
class MyCompanyDatabase:
    table: str = """| Store      | Sales   |
| ---------- | ------- |
| Northeast  | $250    |
| Southeast  | $80     |
| Midwest    | $420    |"""

    def to_markdown(self):
        return self.table


# Mellea also allows you to specify the fields you want to
# include from your class and a corresponding template that
# takes those fields.
@mify(fields_include={"table"}, template="{{ table }}")
class MyOtherCompanyDatabase:
    table: str = """| Store      | Sales   |
| ---------- | ------- |
| Northeast  | $250    |
| Southeast  | $80     |
| Midwest    | $420    |"""


m.query(
    MyOtherCompanyDatabase(), "What were sales for the Northeast branch this month?"
)


# By default, mifying and object will also provide any functions
# of your class/object to models as tools in m functions that support tools.
# The default behavior only includes functions that have docstrings without
# [no-index] in it.
@mify(funcs_include={"from_markdown"})
class MyDocumentLoader:
    def __init__(self) -> None:
        self.content = ""

    @classmethod
    def from_markdown(cls, text: str) -> "MyDocumentLoader":
        doc = MyDocumentLoader()
        # Your parsing functions here.
        doc.content = text
        return doc


# m.transform will be able to call the from_markdown function to return
# the poem as a MyDocumentLoader object.
m.transform(MyDocumentLoader(), "Write a poem.")



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Tests of code under ``mellea.formatters.granite``
"""

# Standard
import copy
import json
import os
import pathlib
from unittest import mock

# Third Party
import openai
import pydantic
import pytest
import requests
import torch
import yaml

# First Party
from mellea.formatters.granite import (
    ChatCompletion,
    ChatCompletionResponse,
    IntrinsicsResultProcessor,
    IntrinsicsRewriter,
)
from mellea.formatters.granite.base import util as base_util
from mellea.formatters.granite.intrinsics import json_util, util as intrinsics_util


def _read_file(name):
    with open(name, encoding="utf-8") as f:
        return f.read()


_TEST_DATA_DIR = pathlib.Path(os.path.dirname(__file__)) / "testdata"

# Location from which our tests download adapters and YAML files
_RAG_INTRINSICS_REPO_NAME = "ibm-granite/granite-lib-rag-r1.0"


_INPUT_JSON_DIR = _TEST_DATA_DIR / "input_json"
_INPUT_YAML_DIR = _TEST_DATA_DIR / "input_yaml"
_INPUT_ARGS_DIR = _TEST_DATA_DIR / "input_args"


class YamlJsonCombo(pydantic.BaseModel):
    """Dataclass that drives configuration for most tests in this file."""

    short_name: str
    """Short name for the test scenario, for printing to the logs."""
    yaml_file: pathlib.Path | None = None
    """Location of local YAML file, or ``None`` to download from remote repo.
    If the file is downloaded, the validator for this field will update this field
    automatically."""
    inputs_file: pathlib.Path
    """Location of local JSON input file."""
    arguments_file: pathlib.Path | None = None
    """Location of local JSON file of arguments, or ``None`` if no arguments."""
    task: str | None = None
    """Name of target task, used for loading adapters. ``None`` means no adapter and
    no inference tests."""
    is_alora: bool = False
    """``True`` to use the activated LoRA variant of the model for inference tests."""
    repo_id: str = _RAG_INTRINSICS_REPO_NAME
    """Repo on Hugging Face Hub from which the adapter for this intrinsic should be
    loaded."""
    revision: str = "main"
    """Revision or branch of the Hugging Face `repo_id`."""
    base_model_id: str = "ibm-granite/granite-4.0-micro"
    """Base model on which the target adapter was trained. Should be small enough to
    run on the CI server."""

    @pydantic.model_validator(mode="after")
    def _maybe_download_yaml(self):
        """
        If YAML file is not provided, download one based on other attributes of this
        object.
        """
        if not self.yaml_file:
            self.yaml_file = intrinsics_util.obtain_io_yaml(
                self.task, self.base_model_id, self.repo_id, revision=self.revision
            )
        return self


_YAML_JSON_COMBOS_LIST = [
    # Short name => YAML file, JSON file, model file, arguments file, is aLoRA
    YamlJsonCombo(
        short_name="answerability_simple",
        inputs_file=_INPUT_JSON_DIR / "simple.json",
        task="answerability",
    ),
    YamlJsonCombo(
        short_name="answerability_extra_params",
        yaml_file=_INPUT_YAML_DIR / "answerability.yaml",
        inputs_file=_INPUT_JSON_DIR / "extra_params.json",
        task=None,  # Fake config, no inference
    ),
    YamlJsonCombo(
        short_name="answerability_answerable",
        inputs_file=_INPUT_JSON_DIR / "answerable.json",
        task="answerability",
    ),
    YamlJsonCombo(
        short_name="answerability_answerable_alora",
        inputs_file=_INPUT_JSON_DIR / "answerable.json",
        task="answerability",
        is_alora=True,
    ),
    YamlJsonCombo(
        short_name="answerability_unanswerable",
        inputs_file=_INPUT_JSON_DIR / "unanswerable.json",
        task="answerability",
    ),
    YamlJsonCombo(
        short_name="instruction",
        yaml_file=_INPUT_YAML_DIR / "instruction.yaml",
        inputs_file=_INPUT_JSON_DIR / "instruction.json",
        arguments_file=_INPUT_ARGS_DIR / "instruction.json",
        task=None,  # Fake config, no model
    ),
    YamlJsonCombo(
        short_name="hallucination_detection",
        inputs_file=_INPUT_JSON_DIR / "hallucination_detection.json",
        task="hallucination_detection",
    ),
    # aLoRA adapter for this intrinsic not currently available
    # YamlJsonCombo(
    #     short_name="hallucination_detection_alora",
    #     inputs_file=_INPUT_JSON_DIR / "hallucination_detection.json",
    #     task="hallucination_detection",
    #     is_alora=True
    # ),
    YamlJsonCombo(
        short_name="query_clarification",
        inputs_file=_INPUT_JSON_DIR / "query_clarification.json",
        task="query_clarification",
    ),
    YamlJsonCombo(
        short_name="query_rewrite",
        inputs_file=_INPUT_JSON_DIR / "query_rewrite.json",
        task="query_rewrite",
    ),
    YamlJsonCombo(
        short_name="requirement_check",
        inputs_file=_INPUT_JSON_DIR / "requirement_check.json",
        arguments_file=_INPUT_ARGS_DIR / "requirement_check.json",
        task="requirement_check",
        # Granite 4.0 adapters not currently available
        repo_id="ibm-granite/rag-intrinsics-lib",
        base_model_id="ibm-granite/granite-3.3-2b-instruct",
    ),
    YamlJsonCombo(
        short_name="requirement_check_alora",
        inputs_file=_INPUT_JSON_DIR / "requirement_check.json",
        arguments_file=_INPUT_ARGS_DIR / "requirement_check.json",
        task="requirement_check",
        is_alora=True,
        # Granite 4.0 adapters not currently available
        repo_id="ibm-granite/rag-intrinsics-lib",
        base_model_id="ibm-granite/granite-3.3-2b-instruct",
    ),
    YamlJsonCombo(
        short_name="uncertainty",
        inputs_file=_INPUT_JSON_DIR / "uncertainty.json",
        task="uncertainty",
        # Granite 4.0 adapters not currently available
        repo_id="ibm-granite/rag-intrinsics-lib",
        base_model_id="ibm-granite/granite-3.3-2b-instruct",
    ),
    YamlJsonCombo(
        short_name="uncertainty_alora",
        inputs_file=_INPUT_JSON_DIR / "uncertainty.json",
        task="uncertainty",
        is_alora=True,
        # Granite 4.0 adapters not currently available
        repo_id="ibm-granite/rag-intrinsics-lib",
        base_model_id="ibm-granite/granite-3.3-2b-instruct",
    ),
    YamlJsonCombo(
        short_name="context_relevance",
        inputs_file=_INPUT_JSON_DIR / "context_relevance.json",
        arguments_file=_INPUT_ARGS_DIR / "context_relevance.json",
        task="context_relevance",
    ),
    YamlJsonCombo(
        short_name="context_relevance_alora",
        inputs_file=_INPUT_JSON_DIR / "context_relevance.json",
        arguments_file=_INPUT_ARGS_DIR / "context_relevance.json",
        task="context_relevance",
        is_alora=True,
    ),
    YamlJsonCombo(
        short_name="citations",
        inputs_file=_INPUT_JSON_DIR / "citations.json",
        task="citations",
    ),
    # aLoRA adapter for this intrinsic not currently available
    # YamlJsonCombo(
    #     short_name="citations_alora",
    #     inputs_file=_INPUT_JSON_DIR / "citations.json",
    #     task="citations",
    #     is_alora=True,
    # ),
    YamlJsonCombo(
        short_name="context-attribution",
        inputs_file=_INPUT_JSON_DIR / "context-attribution.json",
        task="context-attribution",
        repo_id="ibm-granite/granitelib-core-r1.0",
    ),
    # gpt-oss-20b intrinsics (canned output tests only, no inference)
    YamlJsonCombo(
        short_name="gpt_oss_answerability",
        inputs_file=_INPUT_JSON_DIR / "answerable.json",
        task="answerability",
        repo_id="ibm-granite/granite-lib-rag-gpt-oss-r1.0",
        base_model_id="openai/gpt-oss-20b",
    ),
    YamlJsonCombo(
        short_name="gpt_oss_citations",
        inputs_file=_INPUT_JSON_DIR / "citations.json",
        task="citations",
        repo_id="ibm-granite/granite-lib-rag-gpt-oss-r1.0",
        base_model_id="openai/gpt-oss-20b",
    ),
    YamlJsonCombo(
        short_name="gpt_oss_hallucination_detection",
        inputs_file=_INPUT_JSON_DIR / "hallucination_detection.json",
        task="hallucination_detection",
        repo_id="ibm-granite/granite-lib-rag-gpt-oss-r1.0",
        base_model_id="openai/gpt-oss-20b",
    ),
    YamlJsonCombo(
        short_name="gpt_oss_query_rewrite",
        inputs_file=_INPUT_JSON_DIR / "query_rewrite.json",
        task="query_rewrite",
        repo_id="ibm-granite/granite-lib-rag-gpt-oss-r1.0",
        base_model_id="openai/gpt-oss-20b",
    ),
]
_YAML_JSON_COMBOS = {c.short_name: c for c in _YAML_JSON_COMBOS_LIST}


# Base models that are small enough to run locally with transformers
_LOCAL_BASE_MODELS = {
    "ibm-granite/granite-4.0-micro",
    "ibm-granite/granite-3.3-2b-instruct",
}

# All combinations of input and model where a model is present
_YAML_JSON_COMBOS_WITH_MODEL = {
    k: v
    for k, v in _YAML_JSON_COMBOS.items()
    if v.task is not None and v.base_model_id in _LOCAL_BASE_MODELS
}

# All combinations of input and model that are not aLoRA models (includes no model)
_YAML_JSON_COMBOS_NO_ALORA = {
    k: v for k, v in _YAML_JSON_COMBOS.items() if not v.is_alora
}

_YAML_JSON_COMBOS_WITH_LORA_MODEL = {
    k: v for k, v in _YAML_JSON_COMBOS.items() if v.task is not None and not v.is_alora
}

# Combinations suitable for an Ollama backend
_NO_OLLAMA_ADAPTER = {
    # Ollama LoRA adapter not yet available on HF Hub
    "context-attribution"
}
_YAML_JSON_COMBOS_FOR_OLLAMA = {
    k: v
    for k, v in _YAML_JSON_COMBOS.items()
    if v.task is not None
    and not v.is_alora
    and v.base_model_id == "ibm-granite/granite-4.0-micro"
    and k not in _NO_OLLAMA_ADAPTER
}


@pytest.fixture(name="yaml_json_combo", scope="module", params=_YAML_JSON_COMBOS)
def _yaml_json_combo(request: pytest.FixtureRequest) -> YamlJsonCombo:
    """Pytest fixture that allows us to run a given test case repeatedly with multiple
    different combinations of IO configuration and chat completion request.

    Uses the files in ``testdata/input_json`` and ``testdata/input_yaml``.

    Returns test configuration.
    """
    return _YAML_JSON_COMBOS[request.param]


@pytest.fixture(
    name="yaml_json_combo_no_alora", scope="module", params=_YAML_JSON_COMBOS_NO_ALORA
)
def _yaml_json_combo_no_alora(request: pytest.FixtureRequest) -> YamlJsonCombo:
    """Pytest fixture that allows us to run a given test case repeatedly with multiple
    different combinations of IO configuration and chat completion request. Ignores
    model configs that use the aLoRA variant of the model.

    Uses the files in ``testdata/input_json`` and ``testdata/input_yaml``.

    Returns tuple of short name, YAML file, JSON file, model directory, and
    arguments file.
    """
    return _YAML_JSON_COMBOS_NO_ALORA[request.param]


@pytest.fixture(
    name="yaml_json_combo_with_model",
    scope="module",
    params=_YAML_JSON_COMBOS_WITH_MODEL,
)
def _yaml_json_combo_with_model(request: pytest.FixtureRequest) -> YamlJsonCombo:
    """Version of :func:`_yaml_json_combo()` fixture with only the inputs that have
    models.
    """
    return _YAML_JSON_COMBOS_WITH_MODEL[request.param]


@pytest.fixture(
    name="yaml_json_combo_with_lora_model",
    scope="module",
    params=_YAML_JSON_COMBOS_WITH_LORA_MODEL,
)
def _yaml_json_combo_with_lora_model(request: pytest.FixtureRequest) -> YamlJsonCombo:
    """Version of :func:`_yaml_json_combo()` fixture with only the inputs that have
    non-aLoRA models.
    """
    return _YAML_JSON_COMBOS_WITH_LORA_MODEL[request.param]


@pytest.fixture(
    name="yaml_json_combo_for_ollama",
    scope="module",
    params=_YAML_JSON_COMBOS_FOR_OLLAMA,
)
def _yaml_json_combo_for_ollama(request: pytest.FixtureRequest) -> YamlJsonCombo:
    """Version of :func:`_yaml_json_combo()` fixture with only inputs suitable
    for an Ollama backend.
    """
    return _YAML_JSON_COMBOS_FOR_OLLAMA[request.param]


def test_no_orphan_files():
    """Check whether there are input files that aren't used by any test."""
    used_json_files = {t.inputs_file for t in _YAML_JSON_COMBOS.values()}
    all_json_files = list(_INPUT_JSON_DIR.iterdir())
    used_yaml_files = {t.yaml_file for t in _YAML_JSON_COMBOS.values()}
    all_yaml_files = list(_INPUT_YAML_DIR.iterdir())

    for f in all_json_files:
        if f not in used_json_files:
            raise ValueError(
                f"JSON File '{f}' not used. Files are {all_json_files}; "
                f"Used files are {list(used_json_files)}"
            )
    for f in all_yaml_files:
        if f not in used_yaml_files:
            raise ValueError(
                f"YAML File '{f}' not used. Files are {all_yaml_files}; "
                f"Used files are {list(used_yaml_files)}"
            )


def test_read_yaml():
    """Sanity check to verify that reading a model's YAML file from disk works."""
    # Read from local disk
    with open(_INPUT_YAML_DIR / "answerability.yaml", encoding="utf8") as file:
        data = yaml.safe_load(file)
    assert data["model"] is None

    original_data = copy.deepcopy(data)

    # Instantiate directly from dictionary
    IntrinsicsRewriter(config_dict=data)

    # Data shouldn't be modified
    assert original_data == data

    # Manually run through the make_config_dict() function, because apparently users
    # will try to do this.
    data2 = intrinsics_util.make_config_dict(_INPUT_YAML_DIR / "answerability.yaml")
    IntrinsicsRewriter(config_dict=data2)

    # Read from local disk
    IntrinsicsRewriter(config_file=_INPUT_YAML_DIR / "answerability.yaml")

    # Read from Hugging Face hub.
    local_path = intrinsics_util.obtain_io_yaml(
        "answerability", "granite-4.0-micro", _RAG_INTRINSICS_REPO_NAME
    )
    IntrinsicsRewriter(config_file=local_path)


_CANNED_INPUT_EXPECTED_DIR = _TEST_DATA_DIR / "test_canned_input"


def test_canned_input(yaml_json_combo_no_alora):
    """
    Verify that a given combination of chat completion and rewriting config produces
    the expected output
    """
    cfg = yaml_json_combo_no_alora
    if cfg.arguments_file:
        with open(cfg.arguments_file, encoding="utf8") as f:
            transform_kwargs = json.load(f)
    else:
        transform_kwargs = {}

    rewriter = IntrinsicsRewriter(config_file=cfg.yaml_file)

    json_data = _read_file(cfg.inputs_file)
    before = ChatCompletion.model_validate_json(json_data)
    after = rewriter.transform(before, **transform_kwargs)
    after_json = after.model_dump_json(indent=2)

    expected_file = _CANNED_INPUT_EXPECTED_DIR / f"{cfg.short_name}.json"
    with open(expected_file, encoding="utf-8") as f:
        expected_json = f.read()

    print(f"{after_json=}")
    assert after_json == expected_json


@pytest.mark.block_network
def test_openai_compat(yaml_json_combo_no_alora):
    """
    Verify that the dataclasses for intrinsics chat completions can be directly passed
    to the OpenAI Python API without raising parsing errors.
    """
    cfg = yaml_json_combo_no_alora
    if cfg.arguments_file:
        with open(cfg.arguments_file, encoding="utf8") as f:
            transform_kwargs = json.load(f)
    else:
        transform_kwargs = {}

    # Temporary: Use a YAML file from local disk
    rewriter = IntrinsicsRewriter(config_file=cfg.yaml_file)
    json_data = _read_file(cfg.inputs_file)
    before = ChatCompletion.model_validate_json(json_data)
    after = rewriter.transform(before, **transform_kwargs)

    # Create a fake connection to the API so we can use its request validation code.
    # Note that network access is blocked for this test case.
    openai_base_url = "http://example.com:98765/not/a/valid/url"
    openai_api_key = "not_a_valid_api_key"
    client = openai.OpenAI(base_url=openai_base_url, api_key=openai_api_key)

    # OpenAI requires a model name
    before.model = "dummy_model_name"
    after.model = "dummy_model_name"

    # The client should get all the way through validation and fail to connect
    with pytest.raises(openai.APIConnectionError):
        client.chat.completions.create(**(before.model_dump()))
    with pytest.raises(openai.APIConnectionError):
        client.chat.completions.create(**(after.model_dump()))


_CANNED_OUTPUT_MODEL_OUTPUT_DIR = _TEST_DATA_DIR / "test_canned_output/model_output"
_CANNED_OUTPUT_EXPECTED_DIR = _TEST_DATA_DIR / "test_canned_output/expected_result"


def test_canned_output(yaml_json_combo_with_lora_model):
    """
    Verify that the output processing for each model works on previous model outputs
    read from disk. Model outputs are stored in OpenAI format.

    :param yaml_json_combo_no_alora: Same cases as test_canned_input
    """
    # _, yaml_ile, input_file, output_file, expected_file = yaml_output_combo

    # Same cases as test_canned_input
    cfg = yaml_json_combo_with_lora_model

    # Input is input to model, not input to rewriter
    input_file = _CANNED_INPUT_EXPECTED_DIR / f"{cfg.short_name}.json"
    output_file = _CANNED_OUTPUT_MODEL_OUTPUT_DIR / f"{cfg.short_name}.json"
    expected_file = _CANNED_OUTPUT_EXPECTED_DIR / f"{cfg.short_name}.json"

    processor = IntrinsicsResultProcessor(config_file=cfg.yaml_file)
    with open(input_file, encoding="utf-8") as f:
        model_input = ChatCompletion.model_validate_json(f.read())
    with open(output_file, encoding="utf-8") as f:
        model_output = ChatCompletionResponse.model_validate_json(f.read())

    transformed = processor.transform(model_output, model_input)

    # Pull this string out of the debugger to update expected file
    transformed_str = transformed.model_dump_json(indent=4)

    with open(expected_file, encoding="utf-8") as f:
        expected = ChatCompletionResponse.model_validate_json(f.read())
    expected_str = expected.model_dump_json(indent=4)

    # Do an approximate comparison of numeric values.
    # Can't use pytest.approx() because of lists and floats encoded as strings
    transformed_json = _round_floats(json.loads(transformed_str))
    expected_json = _round_floats(json.loads(expected_str))

    assert transformed_json == expected_json


_REPARSE_JSON_DIR = _TEST_DATA_DIR / "test_reparse_json"
_REPARSE_JSON_FILES = [
    name for name in os.listdir(_REPARSE_JSON_DIR) if name.endswith(".json")
]


@pytest.fixture(name="reparse_json_file", scope="module", params=_REPARSE_JSON_FILES)
def _reparse_json_file(request: pytest.FixtureRequest) -> tuple[str, str, str]:
    """Pytest fixture that returns each file in _REPARSE_JSON_DIR in turn"""
    return request.param


def test_reparse_json(reparse_json_file):
    """Ensure that we can reparse JSON data to find position information for
    literals."""
    json_file = _REPARSE_JSON_DIR / reparse_json_file
    json_str = _read_file(json_file)

    parsed_json = json.loads(json_str)
    reparsed_json = json_util.reparse_json_with_offsets(json_str)

    assert json_util.scalar_paths(parsed_json) == json_util.scalar_paths(reparsed_json)


def _round_floats(json_data, num_digits: int = 2):
    """Round all floating-point numbers in a JSON value to facilitate comparisons.

    :param json_data: Arbitrary JSON data.
    :param num_digits: How many decimal points to round to

    :returns: Copy of the input with all floats rounded
    """
    result = copy.deepcopy(json_data)
    for path in json_util.scalar_paths(result):
        value = json_util.fetch_path(result, path)
        if isinstance(value, float):
            json_util.replace_path(result, path, round(value, num_digits))
        elif isinstance(value, str):
            # Test for floating-point number encoded as a string.
            # In Python this test is supposed to use exceptions as control flow.
            try:
                str_as_float = float(value)
                json_util.replace_path(result, path, round(str_as_float, num_digits))
            except ValueError:
                # flow through
                pass

            # Test for JSON object or array encoded as a string
            if value[0] in ("{", "["):
                try:
                    str_as_json = json.loads(value)
                    rounded_json = _round_floats(str_as_json, num_digits)
                    rounded_json_as_str = json.dumps(rounded_json)
                    json_util.replace_path(result, path, rounded_json_as_str)
                except json.JSONDecodeError:
                    # flow through
                    pass
    return result


def test_run_transformers(yaml_json_combo_with_model, gh_run):
    """
    Run the target model end-to-end on transformers.
    """
    # Prevent thrashing when running tests in parallel
    torch.set_num_threads(2)

    cfg = yaml_json_combo_with_model
    if cfg.arguments_file:
        with open(cfg.arguments_file, encoding="utf8") as f:
            transform_kwargs = json.load(f)
    else:
        transform_kwargs = {}

    # Load input request
    with open(cfg.inputs_file, encoding="utf-8") as f:
        model_input = ChatCompletion.model_validate_json(f.read())

    # Download files from Hugging Face Hub
    try:
        lora_dir = intrinsics_util.obtain_lora(
            cfg.task,
            cfg.base_model_id,
            cfg.repo_id,
            revision=cfg.revision,
            alora=cfg.is_alora,
        )
    except requests.exceptions.HTTPError:
        pytest.xfail("Downloads fail on CI server because repo is private")

    # Load IO config YAML for this model
    io_yaml_path = lora_dir / "io.yaml"
    if not os.path.exists(io_yaml_path):
        # Use local files until proper configs are up on Hugging Face
        io_yaml_path = cfg.yaml_file
    rewriter = IntrinsicsRewriter(config_file=io_yaml_path)
    result_processor = IntrinsicsResultProcessor(config_file=io_yaml_path)

    # Prepare inputs for inference
    transformed_input = rewriter.transform(model_input, **transform_kwargs)

    if gh_run:
        pytest.xfail(
            "Skipping end-to-end model evaluation for this test case because it takes "
            "more than 5 seconds. "
            "Mellea's CI fails the entire run without an error message if all 500+ "
            "tests combined take more than 15 minutes to complete. "
            "That works out to 1.8 seconds per test. "
            "Any test that takes more than 5 seconds needs to disable or shortcut "
            "itself during CI, or all of Mellea's development infrastructure will "
            "grind to a halt."
        )

    # Run the model using Hugging Face APIs
    model, tokenizer = base_util.load_transformers_lora(lora_dir)
    generate_input, other_input = (
        base_util.chat_completion_request_to_transformers_inputs(
            transformed_input.model_dump(), tokenizer, model
        )
    )
    responses = base_util.generate_with_transformers(
        tokenizer, model, generate_input, other_input
    )

    # Pull this string out of the debugger to create a fresh model outputs file.
    responses_str = responses.model_dump_json(indent=4)
    print(responses_str[:10000])  # Limit stdout content

    # Output processing
    transformed_responses = result_processor.transform(responses, transformed_input)

    # Pull this string out of the debugger to create a fresh expected file.
    transformed_str = transformed_responses.model_dump_json(indent=4)
    print(transformed_str)

    with open(
        _TEST_DATA_DIR / f"test_run_transformers/{cfg.short_name}.json",
        encoding="utf-8",
    ) as f:
        expected = ChatCompletionResponse.model_validate_json(f.read())
    # expected_str = expected.model_dump_json(indent=4)

    try:
        # Correct for floating point rounding.
        # Can't use pytest.approx() because of lists
        transformed_json = _round_floats(
            json_util.parse_inline_json(transformed_responses.model_dump()),
            num_digits=2,
        )
        expected_json = _round_floats(
            json_util.parse_inline_json(expected.model_dump()), num_digits=2
        )
        if transformed_json != expected_json:
            # Simple comparison failed.
            # Pull out just the content and attempt a more sophisticated comparison
            assert len(transformed_responses.choices) == len(expected.choices)

            for tc, ec in zip(
                transformed_responses.choices, expected.choices, strict=True
            ):
                t_json = json.loads(tc.message.content)
                e_json = json.loads(ec.message.content)

                assert t_json == pytest.approx(e_json, abs=0.1)
    except AssertionError as e:
        # Known intermittent failure under Transformers 5.0
        if cfg.short_name == "hallucination_detection":
            pytest.xfail("Known failure due to Transformers 5.0")
        raise e


def test_run_ollama(yaml_json_combo_for_ollama):
    """
    Run the target model end-to-end with a mock Ollama backend.
    """
    cfg = yaml_json_combo_for_ollama

    # Change base model id to Ollama's version
    if cfg.base_model_id == "ibm-granite/granite-4.0-micro":
        cfg.base_model_id = "granite4:micro"
    else:
        pytest.xfail(f"Unsupported base model: {cfg.base_model_id}")

    if cfg.arguments_file:
        with open(cfg.arguments_file, encoding="utf8") as f:
            transform_kwargs = json.load(f)
    else:
        transform_kwargs = {}

    # Load input request
    with open(cfg.inputs_file, encoding="utf-8") as f:
        model_input = ChatCompletion.model_validate_json(f.read())
    model_input.model = cfg.task

    # Download files from Hugging Face Hub
    try:
        lora_dir = intrinsics_util.obtain_lora(
            cfg.task,
            cfg.base_model_id,
            cfg.repo_id,
            revision=cfg.revision,
            alora=cfg.is_alora,
        )
    except requests.exceptions.HTTPError:
        pytest.xfail("Downloads fail on CI server because repo is private")

    # Load IO config YAML for this model
    io_yaml_path = lora_dir / "io.yaml"
    if not os.path.exists(io_yaml_path):
        # Use local files until proper configs are up on Hugging Face
        io_yaml_path = str(cfg.yaml_file).replace("input_yaml", "input_yaml_ollama")
    rewriter = IntrinsicsRewriter(config_file=io_yaml_path)
    result_processor = IntrinsicsResultProcessor(config_file=io_yaml_path)

    # Prepare inputs for inference
    transformed_input = rewriter.transform(model_input, **transform_kwargs)
    # print(transformed_input.model_dump_json(indent=4))

    # Load a canned model response for the mock Ollama backend
    canned_output_file = _CANNED_OUTPUT_MODEL_OUTPUT_DIR / f"{cfg.short_name}.json"
    with open(canned_output_file, encoding="utf-8") as f:
        mock_response = ChatCompletionResponse.model_validate_json(f.read())

    # Run the model using a mock Ollama backend
    client = openai.OpenAI(
        base_url="http://localhost:55555/v1/", api_key="rag_intrinsics_1234"
    )
    with mock.patch.object(
        client.chat.completions, "create", return_value=mock_response
    ):
        chat_completion = client.chat.completions.create(
            **transformed_input.model_dump()
        )

    # Pull this string out of the debugger to create a fresh model outputs file.
    responses_str = chat_completion.choices[0].model_dump_json(indent=4)
    print("Responses:", responses_str[:100])

    # Output processing
    transformed_responses = result_processor.transform(
        chat_completion, transformed_input
    )

    # Pull this string out of the debugger to create a fresh expected file.
    transformed_str = transformed_responses.model_dump_json(indent=4)
    print("Transformed:", transformed_str[:100])

    with open(
        _TEST_DATA_DIR / f"test_run_ollama/{cfg.short_name}.json", encoding="utf-8"
    ) as f:
        expected = ChatCompletionResponse.model_validate_json(f.read())
    # expected_str = expected.model_dump_json(indent=4)

    # Correct for floating point rounding.
    # Can't use pytest.approx() because of lists
    transformed_json = _round_floats(
        json_util.parse_inline_json(transformed_responses.model_dump()), num_digits=2
    )
    expected_json = _round_floats(
        json_util.parse_inline_json(expected.model_dump()), num_digits=2
    )
    if transformed_json != expected_json:
        # Simple comparison failed.
        # Pull out just the content and attempt a more sophisticated comparison
        assert len(transformed_responses.choices) == len(expected.choices)

        for tc, ec in zip(transformed_responses.choices, expected.choices, strict=True):
            t_json = json.loads(tc.message.content)
            e_json = json.loads(ec.message.content)

            if isinstance(t_json, list):
                # Workaround for pytest bug
                for t_elem, e_elem in zip(t_json, e_json, strict=True):
                    assert t_elem == pytest.approx(e_elem, abs=0.1)
            else:
                if t_json != pytest.approx(e_json, abs=0.1):
                    # Workaround for pytest bug
                    print("Assertion is about to fail.")
                    print(f"   {t_json=}")
                    print(f"   {e_json=}")
                assert t_json == pytest.approx(e_json, abs=0.1)



"""Metrics plugins for recording telemetry data via hooks.

This module contains plugins that hook into the generation pipeline to
automatically record metrics when enabled. Currently includes:

- TokenMetricsPlugin: Records token usage statistics from ModelOutputThunk.usage
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mellea.plugins.base import Plugin
from mellea.plugins.decorators import hook
from mellea.plugins.types import PluginMode

if TYPE_CHECKING:
    from mellea.plugins.hooks.generation import GenerationPostCallPayload


class TokenMetricsPlugin(Plugin, name="token_metrics", priority=50):
    """Records token usage metrics from generation outputs.

    This plugin hooks into the generation_post_call event to automatically
    record token usage metrics when the usage field is populated on
    ModelOutputThunk instances.

    The plugin reads the standardized usage field (OpenAI-compatible format)
    and records metrics following OpenTelemetry Gen-AI semantic conventions.

    Example:
        >>> from mellea.telemetry.metrics_plugins import TokenMetricsPlugin
        >>> from mellea.telemetry.metrics import enable_metrics
        >>>
        >>> enable_metrics()
        >>> with TokenMetricsPlugin():
        ...     result = session.instruct("Hello, world!")
    """

    @hook("generation_post_call", mode=PluginMode.FIRE_AND_FORGET)
    async def record_token_metrics(
        self, payload: GenerationPostCallPayload, context: dict[str, Any]
    ) -> None:
        """Record token metrics after generation completes.

        Args:
            payload: Contains the model_output (ModelOutputThunk) with usage data
            context: Plugin context (unused)
        """
        from mellea.telemetry.metrics import record_token_usage_metrics

        mot = payload.model_output
        if mot.usage is None:
            return

        # Record metrics (no-op if metrics disabled)
        record_token_usage_metrics(
            input_tokens=mot.usage.get("prompt_tokens"),
            output_tokens=mot.usage.get("completion_tokens"),
            model=mot.model or "unknown",
            provider=mot.provider or "unknown",
        )



"""Tests of the code in ``mellea.stdlib.intrinsics.core``"""

import gc
import json
import os
import pathlib

import pytest
import torch

from mellea.backends.huggingface import LocalHFBackend
from mellea.stdlib.components import Document, Message
from mellea.stdlib.components.intrinsic import core
from mellea.stdlib.context import ChatContext
from test.stdlib.components.intrinsic.test_rag import (
    _read_input_json as _read_rag_input_json,
    _read_output_json as _read_rag_output_json,
)

# Skip entire module in CI since all tests are qualitative
pytestmark = [
    pytest.mark.skipif(
        int(os.environ.get("CICD", 0)) == 1,
        reason="Skipping core intrinsic tests in CI - all qualitative tests",
    ),
    pytest.mark.huggingface,
    pytest.mark.requires_gpu,
    pytest.mark.requires_heavy_ram,
    pytest.mark.llm,
]

DATA_ROOT = pathlib.Path(os.path.dirname(__file__)) / "testdata"
"""Location of data files for the tests in this file."""


BASE_MODEL = "ibm-granite/granite-4.0-micro"


@pytest.fixture(name="backend", scope="module")
def _backend():
    """Backend used by the tests in this file. Module-scoped to avoid reloading the 3B model for each test."""
    # Prevent thrashing if the default device is CPU
    torch.set_num_threads(4)

    backend_ = LocalHFBackend(model_id=BASE_MODEL)
    yield backend_

    # Code after yield is cleanup code.
    # Free GPU memory with extreme prejudice.
    del backend_
    gc.collect()
    gc.collect()
    gc.collect()
    torch.cuda.empty_cache()


def _read_input_json(file_name: str):
    """Read test data from JSON and convert to a ChatContext.

    Returns the context and the raw JSON data (for accessing extra fields
    like ``requirement``).
    """
    with open(DATA_ROOT / "input_json" / file_name, encoding="utf-8") as f:
        json_data = json.load(f)

    context = ChatContext()
    for m in json_data["messages"]:
        context = context.add(Message(m["role"], m["content"]))
    return context, json_data


@pytest.mark.qualitative
def test_certainty(backend):
    """Verify that the uncertainty/certainty intrinsic functions properly."""
    context, _ = _read_input_json("uncertainty.json")

    result = core.check_certainty(context, backend)
    assert 0.0 <= result <= 1.0

    result2 = core.check_certainty(context, backend)
    assert 0.0 <= result2 <= 1.0


@pytest.mark.qualitative
def test_requirement_check(backend):
    """Verify that the requirement check intrinsic functions properly."""
    context, json_data = _read_input_json("requirement_check.json")
    requirement = json_data["requirement"]

    result = core.requirement_check(context, backend, requirement)
    assert 0.0 <= result <= 1.0

    result2 = core.requirement_check(context, backend, requirement)
    assert 0.0 <= result2 <= 1.0


@pytest.mark.qualitative
def test_find_context_attributions(backend):
    """Verify that the context-attribution intrinsic functions properly."""
    context, assistant_response, documents = _read_rag_input_json(
        "context-attribution.json"
    )
    expected = _read_rag_output_json("context-attribution.json")

    result = core.find_context_attributions(
        assistant_response, documents, context, backend
    )
    assert result == expected


if __name__ == "__main__":
    pytest.main([__file__])



"""A backend that uses the Huggingface Transformers library.

The purpose of the Hugginface backend is to provide a setting for implementing experimental features. If you want a performance local backend, and do not need experimental features such as Span-based context or ALoras, consider using Ollama backends instead.
"""

from __future__ import annotations

import asyncio
import dataclasses
import datetime
import functools
import json
import threading
from collections.abc import Callable, Coroutine, Sequence
from typing import Any, overload

import llguidance
import llguidance.hf
import llguidance.torch
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers.cache_utils import DynamicCache
from transformers.generation.logits_process import LogitsProcessorList
from transformers.generation.streamers import AsyncTextIteratorStreamer
from transformers.generation.utils import GenerateDecoderOnlyOutput
from transformers.modeling_utils import PreTrainedModel
from transformers.tokenization_utils import PreTrainedTokenizer
from transformers.trainer_utils import set_seed

from ..backends import kv_block_helpers
from ..core import (
    BaseModelSubclass,
    C,
    CBlock,
    Component,
    Context,
    FancyLogger,
    GenerateLog,
    GenerateType,
    ModelOutputThunk,
    Requirement,
)
from ..core.base import AbstractMelleaTool
from ..formatters import ChatFormatter, TemplateFormatter, granite as granite_formatters
from ..helpers import message_to_openai_message, messages_to_docs, send_to_queue
from ..stdlib.components import Intrinsic, Message
from ..stdlib.requirements import ALoraRequirement, LLMaJRequirement
from ..telemetry.backend_instrumentation import (
    instrument_generate_from_raw,
    start_generate_span,
)
from .adapters import (
    AdapterMixin,
    AdapterType,
    IntrinsicAdapter,
    LocalHFAdapter,
    get_adapter_for_intrinsic,
)
from .backend import FormatterBackend
from .cache import Cache, SimpleLRUCache
from .model_ids import ModelIdentifier
from .model_options import ModelOption
from .tools import (
    add_tools_from_context_actions,
    add_tools_from_model_options,
    convert_tools_to_json,
)
from .utils import to_chat, to_tool_calls

"""A configuration type for the unhappy path: Tokenizer * Model * torch device string

Huggingface backends can initialize themselves from a model string if the transformers `Auto*` classes can be used. Therefore, a TransformersTorchConfig usually isn't required. However, sometimes a model needs special care to instantiate properly, or a custom device type needs to bse used. Instead of trying to do a lot of partial magic, we basically have two modaliites: either the constructor can figure out everything from the model_id, or the user has to provide an entire config.
"""
TransformersTorchConfig = tuple[PreTrainedTokenizer, PreTrainedModel, torch.device]

format: None = None  # typing this variable in order to shadow the global format function and ensure mypy checks for errors


@dataclasses.dataclass
class HFAloraCacheInfo:
    """A dataclass for holding a KV cache and associated generation metadata.

    Used by ``LocalHFBackend`` to store intermediate model state that can be
    reused across generation requests via an LRU cache.

    Args:
        kv_cache (DynamicCache | None): The HuggingFace ``DynamicCache`` holding
            precomputed key/value tensors, or ``None`` if not available.
        merged_token_ids (Any): Token IDs corresponding to the cached prefix.
        merged_attention (Any): Attention mask for the cached prefix tokens.
        q_end (int): Index of the last prompt token in the merged token sequence;
            defaults to ``-1``.
        scores (Any): Optional logit scores from the generation step; defaults to
            ``None``.

    Attributes:
        kv_cache (DynamicCache | None): The cached key/value tensors.
        merged_token_ids (Any): Token IDs for the cached prefix.
        merged_attention (Any): Attention mask for the cached prefix.
        q_end (int): End index of the prompt portion in merged token IDs.
        scores (Any): Logit scores from generation, or ``None``.
    """

    kv_cache: DynamicCache | None
    merged_token_ids: Any
    merged_attention: Any
    q_end: int = -1
    scores: Any = None


def _cleanup_kv_cache(cache_info: HFAloraCacheInfo) -> None:
    """Free GPU memory when KV cache is evicted from LRU.

    This function is called by SimpleLRUCache when an entry is evicted.
    It explicitly deletes tensor references and calls torch.cuda.empty_cache()
    to return pooled CUDA memory to the device.

    Args:
        cache_info: The HFAloraCacheInfo being evicted from cache.
    """
    import gc

    if cache_info is None:
        return

    kv = cache_info.kv_cache
    if kv is not None:
        # Delete individual tensors from each layer
        if hasattr(kv, "key_cache"):
            for tensor in kv.key_cache:
                del tensor
            kv.key_cache.clear()
        if hasattr(kv, "value_cache"):
            for tensor in kv.value_cache:
                del tensor
            kv.value_cache.clear()
        del cache_info.kv_cache

    # Delete other tensors
    if cache_info.merged_attention is not None:
        del cache_info.merged_attention

    # Delete score tensors if present
    if cache_info.scores is not None:
        for tensor in cache_info.scores:
            del tensor
        del cache_info.scores

    # Force Python garbage collection and return CUDA memory to device
    gc.collect()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()


# modified from VLLM v0.9.2 code base
# https://github.com/vllm-project/vllm/blob/v0.9.2/vllm/model_executor/guided_decoding/guidance_logits_processors.py
class _GuidanceLogitsProcessor:
    def __init__(self, grammar: str, ll_tokenizer: llguidance.LLTokenizer) -> None:
        self.grammar = grammar
        self.vocab_size: int = ll_tokenizer.vocab_size
        self.ll_tokenizer: llguidance.LLTokenizer = ll_tokenizer
        self.ll_matchers: list[llguidance.LLMatcher] = []
        self.bitmasks: list[torch.Tensor] = []
        self.new_sampling: bool = False
        self.batch_size: int = -1

    def __call__(
        self, batch_input_ids: torch.Tensor, batch_scores: torch.Tensor
    ) -> torch.Tensor:
        i_batch, _ = batch_input_ids.shape
        s_batch, _ = batch_scores.shape
        assert i_batch == s_batch

        # s_batch, s_vocab = batch_scores.shape
        # assert s_vocab == self.vocab_size
        #
        # NOTE: somehow, this does not hold. s_vocab is not same as either of
        # * self._tokenizer._tokenizer.get_vocab_size(with_added_tokens=True) == self.vocab_size == ll_tokenizer.vocab_size
        # * self._tokenizer._tokenizer.get_vocab_size(with_added_tokens=False)

        if self.batch_size != i_batch:
            self.batch_size = i_batch
            self.bitmasks = [
                llguidance.torch.allocate_token_bitmask(1, self.vocab_size)  # type: ignore[attr-defined]
                for _ in range(self.batch_size)
            ]

            self.ll_matchers = [
                llguidance.LLMatcher(self.ll_tokenizer, self.grammar)
                for _ in range(self.batch_size)
            ]

        for input_ids, scores, ll_matcher, bitmask in zip(
            batch_input_ids, batch_scores, self.ll_matchers, self.bitmasks
        ):
            if self.new_sampling and len(input_ids) > 0:
                ll_matcher.consume_token(  # type: ignore[attr-defined]
                    input_ids.tolist()[-1]
                )
                err = ll_matcher.get_error()  # type: ignore[attr-defined]
                if err:
                    FancyLogger.get_logger().warning("Error in LLMatcher: %s", err)

            llguidance.torch.fill_next_token_bitmask(ll_matcher, bitmask, 0)
            llguidance.torch.apply_token_bitmask_inplace(
                scores, bitmask.to(scores.device)
            )  # type: ignore[attr-defined]

        self.new_sampling = True

        return batch_scores


class LocalHFBackend(FormatterBackend, AdapterMixin):
    """The LocalHFBackend uses Huggingface's transformers library for inference, and uses a Formatter to convert `Component`s into prompts. This backend also supports Activated LoRAs (ALoras)](https://arxiv.org/pdf/2504.12397).

    This backend is designed for running an HF model for small-scale inference locally on your machine.

    This backend is NOT designed for inference scaling on CUDA-enabled hardware.

    Args:
        model_id (str | ModelIdentifier): Used to load the model and tokenizer via
            HuggingFace ``Auto*`` classes.
        formatter (ChatFormatter | None): Formatter for rendering components into
            prompts. Defaults to ``TemplateFormatter``.
        use_caches (bool): If ``False``, KV caching is disabled even if a ``Cache``
            is provided.
        cache (Cache | None): Caching strategy; defaults to
            ``SimpleLRUCache(0, on_evict=_cleanup_kv_cache)``.
        custom_config (TransformersTorchConfig | None): Override for
            tokenizer/model/device; if provided, ``model_id`` is not used for loading.
        default_to_constraint_checking_alora (bool): If ``False``, aLoRA constraint
            checking is deactivated; mainly for benchmarking and debugging.
        model_options (dict | None): Default model options for generation requests.

    Attributes:
        to_mellea_model_opts_map (dict): Mapping from HF-specific option names to
            Mellea ``ModelOption`` sentinel keys.
        from_mellea_model_opts_map (dict): Mapping from Mellea sentinel keys to
            HF-specific option names.
    """

    _cached_blocks: dict[str, DynamicCache] = dict()

    def __init__(
        self,
        model_id: str | ModelIdentifier,
        formatter: ChatFormatter | None = None,
        *,
        use_caches: bool = True,
        cache: Cache | None = None,
        custom_config: TransformersTorchConfig | None = None,
        default_to_constraint_checking_alora: bool = True,
        model_options: dict | None = None,
    ):
        """Load model weights from the given model ID, or from a custom config if provided."""
        formatter = (
            formatter if formatter is not None else TemplateFormatter(model_id=model_id)
        )

        super().__init__(model_id, formatter, model_options=model_options)

        # A mapping of common options for this backend mapped to their Mellea ModelOptions equivalent.
        # These are usually values that must be extracted before hand or that are common among backend providers
        self.to_mellea_model_opts_map = {
            "system": ModelOption.SYSTEM_PROMPT,
            "max_new_tokens": ModelOption.MAX_NEW_TOKENS,
            "seed": ModelOption.SEED,
            "tools": ModelOption.TOOLS,
            "stream": ModelOption.STREAM,
        }

        # A mapping of Mellea specific ModelOptions to the specific names for this backend.
        # These options should almost always be a subset of those specified in the `to_mellea_model_opts_map`.
        # Usually, values that are intentionally extracted while prepping for the backend generate call
        # will be omitted here so that they will be removed when model_options are processed
        # for the call to the model.
        self.from_mellea_model_opts_map = {ModelOption.MAX_NEW_TOKENS: "max_new_tokens"}

        self.default_to_constraint_checking_alora = default_to_constraint_checking_alora

        # Either use the custom config or load the model from its model_id
        match model_id:
            case str():
                self._hf_model_id = model_id
            case ModelIdentifier():
                assert model_id.hf_model_name is not None, (
                    "model_id is None. This can also happen if the ModelIdentifier has no hf_model_id name set."
                )
                self._hf_model_id = model_id.hf_model_name
        match custom_config:
            case None:
                # Choose a device.
                self._device = torch.device(
                    "cuda"
                    if torch.cuda.is_available()
                    else "mps"
                    if torch.backends.mps.is_available()
                    else "cpu"
                )
                # Get the model and tokenizer.
                self._model: PreTrainedModel = AutoModelForCausalLM.from_pretrained(
                    self._hf_model_id, device_map=str(self._device)
                )
                self._tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(
                    self._hf_model_id
                )
            case _:
                self._tokenizer, self._model, self._device = custom_config

        self._llguidance_tokenizer: llguidance.LLTokenizer = (
            llguidance.hf.from_tokenizer(self._tokenizer)  # type:ignore
        )
        assert (
            self._llguidance_tokenizer.vocab_size
            == self._tokenizer._tokenizer.get_vocab_size(with_added_tokens=True)
        ), "vocab size mismatch between llguidance and huggingface tokenizers ... wtf?"

        self._use_caches = use_caches
        self._cache = (
            cache
            if cache is not None
            else SimpleLRUCache(0, on_evict=_cleanup_kv_cache)
        )

        # Adapters can be made known to the backend (added) and loaded.
        self._added_adapters: dict[str, LocalHFAdapter] = {}
        self._loaded_adapters: dict[str, LocalHFAdapter] = {}

        self._generation_lock = threading.Lock()
        """Used to force generation requests to be non-concurrent. Necessary for preventing issues with adapters."""

    def _make_dc_cache(self, toks, **model_options):
        dc = DynamicCache()
        with torch.no_grad():
            dc = self._model(
                toks["input_ids"].to(self._device),
                attention_mask=toks["attention_mask"].to(self._device),
                past_key_values=dc,
                **model_options,
            ).past_key_values
        return dc

    async def _generate_from_context(
        self,
        action: Component[C] | CBlock,
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> tuple[ModelOutputThunk[C], Context]:
        """Generate a completion for ``action`` given ``ctx`` using the HuggingFace model.

        Automatically routes ``Requirement`` and ``Intrinsic`` actions to their
        corresponding aLoRA adapters when available.

        Args:
            action (Component[C] | CBlock): The component or content block to generate
                a completion for.
            ctx (Context): The current generation context (must be a chat context).
            format (type[BaseModelSubclass] | None): Optional Pydantic model class for
                structured/constrained output decoding via llguidance.
            model_options (dict | None): Per-call model options that override the
                backend's defaults.
            tool_calls (bool): If ``True``, expose available tools to the model and
                parse tool-call responses.

        Returns:
            tuple[ModelOutputThunk[C], Context]: A thunk holding the (lazy) model output
                and an updated context that includes ``action`` and the new output.
        """
        span = start_generate_span(
            backend=self, action=action, ctx=ctx, format=format, tool_calls=tool_calls
        )
        await self.do_generate_walk(action)

        # Upsert model options.
        model_opts = self._simplify_and_merge(model_options)

        # Requirements can be automatically rerouted to a requirement adapter.
        if isinstance(action, Requirement):
            # See docs/dev/requirement_aLoRA_rerouting.md
            reroute_to_alora = self.default_to_constraint_checking_alora
            adapter_name = "requirement_check"

            if isinstance(action, ALoraRequirement):
                reroute_to_alora = True
                adapter_name = action.intrinsic_name
                alora_action = action
            else:
                assert action.description is not None, (
                    "must have a description when generating from a requirement"
                )
                alora_action = ALoraRequirement(action.description, adapter_name)

            # Check if a requirement_check (or AloraRequirement specified) adapter
            # exists.
            alora_req_adapter = get_adapter_for_intrinsic(
                adapter_name, [AdapterType.ALORA], self._added_adapters
            )
            if alora_req_adapter is None:
                # Log a warning if using an AloraRequirement but no adapter fit.
                if reroute_to_alora and isinstance(action, ALoraRequirement):
                    FancyLogger.get_logger().warning(
                        f"attempted to use an AloraRequirement but backend {self} doesn't have the specified adapter added {adapter_name}; defaulting to regular generation"
                    )
                reroute_to_alora = False

            if issubclass(type(action), LLMaJRequirement):
                reroute_to_alora = False

            if reroute_to_alora:
                # Keep the alora requirement handling separate for now.
                mot = await self._generate_from_intrinsic(
                    alora_action, ctx, model_options=model_opts
                )
                # Store span for telemetry
                if span is not None:
                    mot._meta["_telemetry_span"] = span
                return mot, ctx.add(alora_action).add(mot)

        elif isinstance(action, Intrinsic):
            mot = await self._generate_from_intrinsic(
                action, ctx, model_options=model_opts
            )
            # Store span for telemetry
            if span is not None:
                mot._meta["_telemetry_span"] = span
            return mot, ctx.add(action).add(mot)

        mot = await self._generate_from_context_standard(
            action, ctx, _format=format, model_options=model_opts, tool_calls=tool_calls
        )

        # Store span in metadata for post_processing to record telemetry
        if span is not None:
            mot._meta["_telemetry_span"] = span

        return mot, ctx.add(action).add(mot)

    def _generate_with_adapter_lock(
        self, adapter_name: str, generate_func: Callable, *args, **kwargs
    ):
        """Helper function for ensuring exclusive generation when adapters are present. Necessary to prevent generating with incorrect weights."""
        with self._generation_lock:
            if adapter_name != "":
                self.load_adapter(adapter_name)
                self._model.set_adapter(adapter_name)
            else:
                try:
                    # `._model.disable_adapters()` doesn't seem to actually disable them or
                    # remove them from the model's list of `.active_adapters()`.
                    self._model.set_adapter([])
                except ValueError as e:
                    # If no weights have been loaded, the model will raise a ValueError:
                    # `ValueError("No adapter loaded. Please load an adapter first.")`
                    if "No adapter loaded" in str(e):
                        pass
                    else:
                        raise e

            _assert_correct_adapters(adapter_name, self._model)
            out = generate_func(*args, **kwargs)
            _assert_correct_adapters(adapter_name, self._model)
            return out

    async def _generate_from_intrinsic(
        self, action: Intrinsic, ctx: Context, *, model_options: dict[str, Any]
    ) -> ModelOutputThunk:
        if not ctx.is_chat_context:
            raise Exception("Does not yet support non-chat contexts.")

        if len(model_options.items()) > 0:
            FancyLogger.get_logger().info(
                "passing in model options when generating with an adapter; some model options may be overwritten / ignored"
            )

        linearized_ctx = ctx.view_for_generation()
        assert linearized_ctx is not None, (
            "If ctx.is_chat_context, then the context should be linearizable."
        )
        ctx_as_message_list: list[Message] = self.formatter.to_chat_messages(
            linearized_ctx
        )

        # NOTE: Explicitly do not add the action to the context here. Intrinsics modify the context
        #       through their rewriters.

        conversation: list[dict] = []
        system_prompt = model_options.get(ModelOption.SYSTEM_PROMPT, "")
        if system_prompt != "":
            conversation.append({"role": "system", "content": system_prompt})

        conversation.extend([message_to_openai_message(m) for m in ctx_as_message_list])

        docs = messages_to_docs(ctx_as_message_list)

        seed = model_options.get(ModelOption.SEED, None)
        if seed is not None:
            set_seed(seed)

        if model_options.get(ModelOption.STREAM, None) is not None:
            # Intrinsics don't support streaming because of their post-processing step.
            raise Exception("Intrinsics do not support streaming.")

        adapter = get_adapter_for_intrinsic(
            action.intrinsic_name, action.adapter_types, self._added_adapters
        )
        if adapter is None:
            raise ValueError(
                f"backend ({self}) has no adapter for processing intrinsic: {action.intrinsic_name}"
            )

        # TODO: Code below this point is mostly specific to RagIntrinsics
        #       It should be refactored into a specific adapter.transform() function.
        assert isinstance(adapter, IntrinsicAdapter), (
            "currently Mellea only supports IntrinsicAdapters and Intrinsics"
        )

        intrinsic_config = adapter.config
        assert intrinsic_config is not None

        rewriter = granite_formatters.IntrinsicsRewriter(
            config_dict=intrinsic_config, model_name=adapter.name
        )
        result_processor = granite_formatters.IntrinsicsResultProcessor(
            config_dict=intrinsic_config
        )

        # Convert our conversation into a proper chat completions dict.
        # [{role: user, content: Hello}, {...}] -> {messages: [{role:user,...}, ...], model:..., ...}
        request_json: dict = {
            "messages": conversation,
            "extra_body": {"documents": docs},
        }

        # Convert other parameters from Mellea proprietary format to standard format.
        for model_option in model_options:
            if model_option == ModelOption.TEMPERATURE:
                request_json["temperature"] = model_options[model_option]

        rewritten = rewriter.transform(request_json, **action.intrinsic_kwargs)

        # TODO: Handle caching here. granite_formatters doesn't tell us what changed,
        #       so we will have to invalidate the cache on our side. This requires
        #       us having specific caching for each Component/Message.

        generate_input, other_input = (
            granite_formatters.base.util.chat_completion_request_to_transformers_inputs(  # type: ignore
                rewritten, self._tokenizer, self._model
            )
        )

        chat_response = asyncio.to_thread(
            self._generate_with_adapter_lock,
            adapter.qualified_name,
            granite_formatters.base.util.generate_with_transformers,  # type: ignore
            # Passed as args/kwargs to generate.
            self._tokenizer,
            self._model,
            generate_input,
            other_input,
        )

        output = ModelOutputThunk(None)
        output._start = datetime.datetime.now()
        output._context = ctx.view_for_generation()
        output._action = action
        output._model_options = model_options

        # Add another step to the processing function.
        async def granite_formatters_processing(
            mot: ModelOutputThunk,
            chunk: granite_formatters.ChatCompletionResponse,
            rewritten: granite_formatters.ChatCompletion,
            result_processor: granite_formatters.IntrinsicsResultProcessor,
            input_ids,
        ):
            try:
                res = result_processor.transform(chunk, rewritten)  # type: ignore
            except json.JSONDecodeError as e:
                raise Exception(f"Intrinsic did not return a JSON: {chunk}") from e

            # TODO: If we want to support caches, we need to get the GenerateDecoderOnlyOutput. This means we
            #       probably need to break out the pieces from `generate_with_transformers`.
            # processing expects a str or a GenerateDecoderOnlyOutput. Extract the str.
            return await self.processing(
                mot, res.choices[0].message.content, input_ids=input_ids
            )

        output._process = functools.partial(
            granite_formatters_processing,
            rewritten=rewritten,
            result_processor=result_processor,
            input_ids=generate_input["input_tokens"],
        )

        # TODO: Post-processing should release the lock for this generation.
        output._post_process = functools.partial(
            self.post_processing,
            conversation=conversation,
            input_ids=generate_input["input_tokens"],
            _format=None,
            tool_calls=False,
            tools={},
            seed=seed,
        )

        try:
            # To support lazy computation, will need to remove this create_task and store just the unexecuted coroutine.
            # We can also support synchronous calls by adding a flag and changing this ._generate function.

            # This function should always be called from a running event loop so we don't have to worry about
            # scheduling the task to a specific event loop here.
            output._generate = asyncio.create_task(
                send_to_queue(chat_response, output._async_queue)  # type: ignore
            )
            output._generate_type = GenerateType.ASYNC
        except RuntimeError as e:
            # Most likely cause is running this function without an event loop present.
            raise e

        return output

    # TODO make this async.
    def _make_merged_kv_cache(
        self,
        linearized_ctx: list[Component | CBlock | ModelOutputThunk],
        ctx_as_conversation: Any,
        model_options: Any,
        tools: Any,
    ):
        # Explanation for code blocks inside of use_kv_cache checks:
        # 1. cache every CBlock that is marked with `cache=True` and store in _cached_blocks.
        # 2. Mark each "hit" by adding the string (tokenized?) value to `cached_block_keys`.
        # 3. apply the chat template (without?) tokenization
        # 4. split on cache hits
        # 5. prefill + smash together everything.
        # 6. generate

        # 1. cache every CBlock that is marked with `cache=True` and store in _cached_blocks.
        # AND
        # 2. Mark each "hit" by adding the string (tokenized?) value to `cached_block_keys`.
        cached_block_keys = []
        for c in linearized_ctx:
            match c:
                case CBlock() if c.cache:
                    assert c.value is not None
                    if c.value in self._cached_blocks:
                        FancyLogger.get_logger().info(
                            f"KV CACHE HIT for: {hash(c.value)} ({c.value[:3]}..{c.value[-3:]})"  # type: ignore
                        )
                    else:
                        FancyLogger.get_logger().debug(
                            f"HF backend is caching a CBlock with hashed contents: {hash(c.value)} ({c.value[:3]}..{c.value[-3:]})"
                        )
                        tokens = self._tokenizer(c.value, return_tensors="pt")
                        self._cached_blocks[c.value] = self._make_dc_cache(tokens)
                        cached_block_keys.append(c.value)
                case _:
                    continue

        # 3. apply the chat template WITHOUT tokenization.
        # Doing this without tokenization and then gluing together the tokens is necessary because
        # things that KV cache together must tokenize together.
        input_text = self._tokenizer.apply_chat_template(  # type: ignore
            ctx_as_conversation,
            tools=convert_tools_to_json(tools),  # type: ignore
            **self._make_backend_specific_and_remove(model_options),
            tokenize=False,
        )

        # 4. split the input_text back up again, re-using DC where it exists.
        str_parts = []
        tok_parts = []
        dc_parts = []
        current_suffix = input_text
        for key in cached_block_keys:
            assert key is not None, (
                "Some input CBlock must not have bee ncomputed yet? The error comes far before this line."
            )
            assert key in current_suffix, (
                "Could happen but would be rare. related to the other assert in this block."
            )
            parts = current_suffix.split(key)  # type: ignore
            assert len(parts) == 2, (
                "Known issue: cached substring might occur more than once. We need to handle this situation earlier. Notice if this happens and keep a count."
            )
            prefix, suffix = parts
            # Add the prefix, if any, to str+tok+dc parts.
            if prefix != "":
                FancyLogger.get_logger().debug(
                    f"Doing a forward pass on uncached block which is prefix to a cached CBlock: {prefix[:3]}.{len(prefix)}.{prefix[-3:]}"
                )
                str_parts.append(prefix)
                tok_parts.append(self._tokenizer(prefix, return_tensors="pt"))
                dc_parts.append(self._make_dc_cache(tok_parts[-1]))
            # Add the cached CBlock to str+tok+dc parts.
            FancyLogger.get_logger().debug(
                f"Replacing a substring with previously computed/retrieved cache with hahs value {hash(key)} ({key[:3]}..{key[-3:]})"
            )
            # str_parts.append(key)
            # tok_parts.append(self._tokenizer(key, return_tensors="pt"))
            # dc_parts.append(self._make_dc_cache(tok_parts[-1])) # TODO this is wrong.
            str_parts.append(key)
            tok_parts.append(self._tokenizer(key, return_tensors="pt"))
            dc_parts.append(self._cached_blocks[key])
            # set the suffix for the next loop iteration.
            current_suffix = suffix
        # "base" case: the final suffix.
        if current_suffix != "":
            FancyLogger.get_logger().debug(  # type: ignore
                f"Doing a forward pass on final suffix, an uncached block: {current_suffix[:3]}.{len(current_suffix)}.{current_suffix[-3:]}"  # type: ignore
            )  # type: ignore
            str_parts.append(current_suffix)
            tok_parts.append(self._tokenizer(current_suffix, return_tensors="pt"))
            dc_parts.append(self._make_dc_cache(tok_parts[-1]))

        # Smash together the caches, the input_ids, and the attention masks.
        assert "".join(str_parts) == input_text, (
            "Should've ended up with the same input text!"
        )
        input_ids = torch.cat([toks["input_ids"] for toks in tok_parts], dim=1)
        attention_mask = torch.cat(
            [toks["attention_mask"] for toks in tok_parts], dim=1
        )
        assert input_ids.shape == attention_mask.shape
        merged_cache: DynamicCache = kv_block_helpers.merge_dynamic_caches(dc_parts)
        # TODO: also assert that the merged cached is the correct shape given the input_ids and attention_mask shapes.
        # rewind merged cache by 1 for safety.
        merged_cache.crop(-1)  # type: ignore
        # Return the merged cache.
        return input_text, input_ids, merged_cache, attention_mask

    async def _generate_from_context_with_kv_cache(
        self,
        action: Component[C] | CBlock,
        ctx: Context,
        *,
        _format: type[BaseModelSubclass] | None = None,
        model_options: dict[str, Any],
        tool_calls: bool = False,
    ) -> ModelOutputThunk[C]:
        # Construct input.
        # If the Context is a ChatHistory then we will pretty-print each content as a message and then use apply_chat_template.
        # Otherwise, we will linearize the context and treat it as a raw input.
        if ctx.is_chat_context:
            system_prompt = model_options.get(ModelOption.SYSTEM_PROMPT, None)
            ctx_as_chat = to_chat(action, ctx, self.formatter, system_prompt)

            # Append tool call information if applicable.
            tools: dict[str, AbstractMelleaTool] = dict()
            if tool_calls:
                if _format:
                    FancyLogger.get_logger().warning(
                        f"Tool calling typically uses constrained generation, but you have specified a `format` in your generate call. NB: tool calling is superseded by format; we will NOT call tools for your request: {action}"
                    )
                else:
                    add_tools_from_model_options(tools, model_options)
                    add_tools_from_context_actions(
                        tools, ctx.actions_for_available_tools()
                    )

                    # Add the tools from the action for this generation last so that
                    # they overwrite conflicting names.
                    add_tools_from_context_actions(tools, [action])
                FancyLogger.get_logger().info(f"Tools for call: {tools.keys()}")

            seed = model_options.get(ModelOption.SEED, None)
            if seed is not None:
                set_seed(seed)

            format_kwargs = {}
            if _format:
                schema: dict[str, Any] = _format.model_json_schema()
                grammar: str = llguidance.LLMatcher.grammar_from_json_schema(
                    schema, defaults={"whitespace_flexible": False}
                )
                logits_processor = _GuidanceLogitsProcessor(
                    grammar, self._llguidance_tokenizer
                )
                format_kwargs["logits_processor"] = LogitsProcessorList(
                    [logits_processor]
                )

            streaming_kwargs = {}
            streamer = None
            stream = model_options.get(ModelOption.STREAM, False)
            if stream:
                try:
                    # HuggingFace uses a streaming interface that you pass to the generate call.
                    # Must be called from a running event loop. This should always be the case given the same
                    # requirement of the ._generate function below.
                    streamer = AsyncTextIteratorStreamer(
                        self._tokenizer,  # type: ignore
                        skip_prompt=True,
                        skip_special_tokens=True,
                    )
                    streaming_kwargs["streamer"] = streamer
                except RuntimeError as e:
                    # Most likely cause is creating this object without an event loop present.
                    raise e

            # Create a separate thread to handle the processing. Make it awaitable
            # for non-streaming cases and to get the final output.
            # Details: https://huggingface.co/docs/transformers/en/internal/generation_utils#transformers.AsyncTextIteratorStreamer

            # Filter out chat template-only options before passing to generate()
            generate_options = self._filter_chat_template_only_options(model_options)

            linearized_ctx = ctx.view_for_generation()
            assert linearized_ctx is not None
            _input_text, input_ids, merged_cache, attention_mask = (
                self._make_merged_kv_cache(
                    linearized_ctx=linearized_ctx,
                    ctx_as_conversation=ctx_as_chat,
                    model_options=model_options,
                    tools=tools,
                )
            )

            chat_response = asyncio.to_thread(
                self._generate_with_adapter_lock,
                "",  # Empty for no adapters.
                self._model.generate,  # type: ignore
                # Passed as args/kwargs to generate.
                input_ids.to(self._device),
                use_cache=True,
                past_key_values=merged_cache,
                attention_mask=attention_mask.to(self._device),
                return_dict_in_generate=True,
                output_scores=True,
                **self._make_backend_specific_and_remove(generate_options),
                **streaming_kwargs,  # type: ignore
                **format_kwargs,  # type: ignore
            )

            output = ModelOutputThunk(None)
            output._start = datetime.datetime.now()
            output._context = ctx.view_for_generation()
            output._action = action
            output._model_options = model_options

            # Processing functions only pass the ModelOutputThunk (and current chunk of response). Bind the other vars necessary for
            # each processing step.
            output._process = functools.partial(self.processing, input_ids=input_ids)
            output._post_process = functools.partial(
                self.post_processing,
                conversation=ctx_as_chat,
                input_ids=input_ids,
                _format=_format,
                tool_calls=tool_calls,
                tools=tools,
                seed=seed,
            )

            try:
                # To support lazy computation, will need to remove this create_task and store just the unexecuted coroutine.
                # We can also support synchronous calls by adding a flag and changing this ._generate function.

                response: AsyncTextIteratorStreamer | Coroutine = chat_response
                if stream and streamer is not None:
                    # For streaming, we want to pass the AsyncIterator to the function. Unlike other backends,
                    # this isn't returned by the chat_response coroutine. So we handle it here.
                    response = streamer

                    # Since the async iterator isn't returned by the chat_response coroutine, we have to create a separate
                    # task for it here so that it runs in the background. Attach it to the ModelOutputThunk.
                    output._generate_extra = asyncio.create_task(chat_response)

                # This function should always be called from a running event loop so we don't have to worry about
                # scheduling the task to a specific event loop here.
                output._generate = asyncio.create_task(
                    send_to_queue(response, output._async_queue)  # type: ignore
                )
                output._generate_type = GenerateType.ASYNC
            except RuntimeError as e:
                # Most likely cause is running this function without an event loop present.
                raise e

            return output

        else:
            raise Exception("Does not yet support non-chat contexts.")

    async def _generate_from_context_standard(
        self,
        action: Component | CBlock,
        ctx: Context,
        *,
        _format: type[BaseModelSubclass] | None = None,
        model_options: dict[str, Any],
        tool_calls: bool = False,
    ) -> ModelOutputThunk:
        # Construct input.
        # If the Context is a ChatHistory then we will pretty-print each content as a message and then use apply_chat_template.
        # Otherwise, we will linearize the context and treat it as a raw input.
        if ctx.is_chat_context:
            system_prompt = model_options.get(ModelOption.SYSTEM_PROMPT, None)
            ctx_as_chat = to_chat(action, ctx, self.formatter, system_prompt)

            # Append tool call information if applicable.
            tools: dict[str, AbstractMelleaTool] = dict()
            if tool_calls:
                if _format:
                    FancyLogger.get_logger().warning(
                        f"Tool calling typically uses constrained generation, but you have specified a `format` in your generate call. NB: tool calling is superseded by format; we will NOT call tools for your request: {action}"
                    )
                else:
                    add_tools_from_model_options(tools, model_options)
                    add_tools_from_context_actions(
                        tools, ctx.actions_for_available_tools()
                    )

                    # Add the tools from the action for this generation last so that
                    # they overwrite conflicting names.
                    add_tools_from_context_actions(tools, [action])
                FancyLogger.get_logger().info(f"Tools for call: {tools.keys()}")

            seed = model_options.get(ModelOption.SEED, None)
            if seed is not None:
                set_seed(seed)

            input_ids = self._tokenizer.apply_chat_template(  # type: ignore
                ctx_as_chat,
                tools=convert_tools_to_json(tools),  # type: ignore
                add_generation_prompt=True,  # If we change this, must modify huggingface granite guardian.
                return_tensors="pt",
                **self._make_backend_specific_and_remove(model_options),
            ).to(self._device)  # type: ignore

            format_kwargs = {}
            if _format:
                schema: dict[str, Any] = _format.model_json_schema()
                grammar: str = llguidance.LLMatcher.grammar_from_json_schema(
                    schema, defaults={"whitespace_flexible": False}
                )
                logits_processor = _GuidanceLogitsProcessor(
                    grammar, self._llguidance_tokenizer
                )
                format_kwargs["logits_processor"] = LogitsProcessorList(
                    [logits_processor]
                )

            streaming_kwargs = {}
            streamer = None
            stream = model_options.get(ModelOption.STREAM, False)
            if stream:
                try:
                    # HuggingFace uses a streaming interface that you pass to the generate call.
                    # Must be called from a running event loop. This should always be the case given the same
                    # requirement of the ._generate function below.
                    streamer = AsyncTextIteratorStreamer(
                        self._tokenizer,  # type: ignore
                        skip_prompt=True,
                        skip_special_tokens=True,
                    )
                    streaming_kwargs["streamer"] = streamer
                except RuntimeError as e:
                    # Most likely cause is creating this object without an event loop present.
                    raise e

            # Create a separate thread to handle the processing. Make it awaitable
            # for non-streaming cases and to get the final output.
            # Details: https://huggingface.co/docs/transformers/en/internal/generation_utils#transformers.AsyncTextIteratorStreamer

            # Filter out chat template-only options before passing to generate()
            generate_options = self._filter_chat_template_only_options(model_options)

            chat_response = asyncio.to_thread(
                self._generate_with_adapter_lock,
                "",  # Empty for no adapters.
                self._model.generate,  # type: ignore
                # Passed as args/kwargs to generate.
                input_ids,
                return_dict_in_generate=True,
                use_cache=self._use_caches,  # Only create KV cache if caching is enabled
                **self._make_backend_specific_and_remove(generate_options),
                **streaming_kwargs,  # type: ignore
                **format_kwargs,  # type: ignore
            )

            output = ModelOutputThunk(None)
            output._start = datetime.datetime.now()
            output._context = ctx.view_for_generation()
            output._action = action
            output._model_options = model_options

            # Processing functions only pass the ModelOutputThunk (and current chunk of response). Bind the other vars necessary for
            # each processing step.
            output._process = functools.partial(self.processing, input_ids=input_ids)
            output._post_process = functools.partial(
                self.post_processing,
                conversation=ctx_as_chat,
                input_ids=input_ids,
                _format=_format,
                tool_calls=tool_calls,
                tools=tools,
                seed=seed,
            )

            try:
                # To support lazy computation, will need to remove this create_task and store just the unexecuted coroutine.
                # We can also support synchronous calls by adding a flag and changing this ._generate function.

                response: AsyncTextIteratorStreamer | Coroutine = chat_response
                if stream and streamer is not None:
                    # For streaming, we want to pass the AsyncIterator to the function. Unlike other backends,
                    # this isn't returned by the chat_response coroutine. So we handle it here.
                    response = streamer

                    # Since the async iterator isn't returned by the chat_response coroutine, we have to create a separate
                    # task for it here so that it runs in the background. Attach it to the ModelOutputThunk.
                    output._generate_extra = asyncio.create_task(chat_response)

                # This function should always be called from a running event loop so we don't have to worry about
                # scheduling the task to a specific event loop here.
                output._generate = asyncio.create_task(
                    send_to_queue(response, output._async_queue)  # type: ignore
                )
                output._generate_type = GenerateType.ASYNC
            except RuntimeError as e:
                # Most likely cause is running this function without an event loop present.
                raise e

            return output

        else:
            raise Exception("Does not yet support non-chat contexts.")

    async def processing(
        self, mot: ModelOutputThunk, chunk: str | GenerateDecoderOnlyOutput, input_ids
    ):
        """Accumulate decoded text from a streaming chunk or full generation output.

        For streaming responses the chunk is an already-decoded string from
        ``AsyncTextIteratorStreamer``; for non-streaming it is a
        ``GenerateDecoderOnlyOutput`` that is decoded here.

        Args:
            mot (ModelOutputThunk): The output thunk being populated.
            chunk (str | GenerateDecoderOnlyOutput): A decoded text chunk (streaming)
                or a full HuggingFace generation output object (non-streaming).
            input_ids: The prompt token IDs used for decoding; required to slice off
                the prompt portion from the generated sequences.
        """
        if mot._underlying_value is None:
            mot._underlying_value = ""

        # Because we use the AsyncTextIteratorStreamer, streaming responses are of type str;
        # and already decoded.
        if isinstance(chunk, str):
            mot._underlying_value += chunk
        elif isinstance(chunk, GenerateDecoderOnlyOutput):
            # Otherwise, it's a non-streaming request. Decode it here.
            mot._meta["hf_output"] = chunk
            mot._underlying_value += self._tokenizer.decode(
                chunk.sequences[0, input_ids.shape[1] :], skip_special_tokens=True
            )

    async def post_processing(
        self,
        mot: ModelOutputThunk,
        conversation: list[dict],
        _format: type[BaseModelSubclass] | None,
        tool_calls: bool,
        tools: dict[str, AbstractMelleaTool],
        seed,
        input_ids,
    ):
        """Finalize the output thunk after HuggingFace generation completes.

        Stores the KV cache for future reuse, parses tool calls if applicable,
        records token usage metrics, emits telemetry, and attaches the generate log.

        Args:
            mot (ModelOutputThunk): The output thunk to finalize.
            conversation (list[dict]): The chat conversation sent to the model,
                used for logging.
            _format (type[BaseModelSubclass] | None): The structured output format
                class used during generation, if any.
            tool_calls (bool): Whether tool calling was enabled for this request.
            tools (dict[str, AbstractMelleaTool]): Available tools, keyed by name.
            seed: The random seed used during generation, or ``None``.
            input_ids: The prompt token IDs; used to compute token counts and for
                KV cache bookkeeping.
        """
        if mot._meta.get("hf_output", None) is None:
            if mot._generate_extra is not None:
                full_output = await mot._generate_extra
                assert isinstance(full_output, GenerateDecoderOnlyOutput)
                mot._meta["hf_output"] = full_output

        # The ModelOutputThunk must be computed by this point.
        assert mot.value is not None

        # Store KV cache in LRU separately (not in mot._meta) to enable proper cleanup on eviction.
        # This prevents GPU memory from being held by ModelOutputThunk references.
        hf_output = mot._meta.get("hf_output", None)
        if (
            self._use_caches
            and isinstance(hf_output, GenerateDecoderOnlyOutput)
            and (hf_output.past_key_values is not None or hf_output.scores is not None)
        ):
            output_complete = hf_output.sequences[0]
            kv_cache: DynamicCache | None = hf_output.past_key_values  # type: ignore

            cache_info = HFAloraCacheInfo(
                kv_cache=kv_cache,
                merged_token_ids=output_complete,
                merged_attention=torch.ones_like(output_complete).to(self._device),
                q_end=len(input_ids[0]),  # type: ignore
                scores=hf_output.scores,
            )

            cache_key = id(mot.value)
            self.cache_put(cache_key, cache_info)

            # Clear KV cache and scores from HF output - they're now owned by the LRU cache
            hf_output.past_key_values = None
            hf_output.scores = None

        # Only scan for tools if we are not doing structured output and tool calls were provided to the model.
        if _format is None and tool_calls:
            mot.tool_calls = to_tool_calls(tools, mot.value)

        assert mot._action is not None, (
            "ModelOutputThunks should have their action assigned during generation"
        )
        assert mot._model_options is not None, (
            "ModelOutputThunks should have their model_opts assigned during generation"
        )

        span = mot._meta.get("_telemetry_span")

        # Derive token counts from the output sequences (HF models have no usage object).
        hf_output = mot._meta.get("hf_output")
        n_prompt, n_completion = None, None
        if isinstance(hf_output, GenerateDecoderOnlyOutput):
            try:
                if input_ids is not None and hf_output.sequences is not None:
                    n_prompt = input_ids.shape[1]
                    n_completion = hf_output.sequences[0].shape[0] - n_prompt
            except Exception:
                pass

        if n_prompt is not None and n_completion is not None:
            mot.usage = {
                "prompt_tokens": n_prompt,
                "completion_tokens": n_completion,
                "total_tokens": n_prompt + n_completion,
            }

        # Populate model and provider metadata
        if hasattr(self.model_id, "hf_model_name"):
            mot.model = str(self.model_id.hf_model_name)  # type: ignore
        else:
            mot.model = str(self.model_id)
        mot.provider = "huggingface"

        # Record tracing if span exists
        if span is not None:
            from ..telemetry import end_backend_span
            from ..telemetry.backend_instrumentation import (
                record_response_metadata,
                record_token_usage,
            )

            if isinstance(hf_output, GenerateDecoderOnlyOutput):
                record_response_metadata(span, hf_output)
                if mot.usage:
                    record_token_usage(span, mot.usage)

            # Close the span now that async operation is complete
            end_backend_span(span)
            # Clean up span reference
            del mot._meta["_telemetry_span"]

        # When caching is disabled, clear hf_output from meta to free GPU memory.
        # The sequences tensor is on GPU and accumulates if not cleared.
        if not self._use_caches and isinstance(
            mot._meta.get("hf_output"), GenerateDecoderOnlyOutput
        ):
            import gc

            hf_out = mot._meta["hf_output"]
            if hasattr(hf_out, "sequences") and hf_out.sequences is not None:
                del hf_out.sequences
            if hasattr(hf_out, "scores") and hf_out.scores is not None:
                del hf_out.scores
            del mot._meta["hf_output"]

            # Force Python GC and return CUDA memory to device
            gc.collect()
            torch.cuda.empty_cache()

        # Generate the log for this ModelOutputThunk.
        generate_log = GenerateLog()
        generate_log.prompt = conversation
        generate_log.backend = f"hf::{self.model_id!s}"
        generate_log.model_options = mot._model_options
        generate_log.date = datetime.datetime.now()
        generate_log.model_output = mot.value
        generate_log.extra = {
            "format": _format,
            "tools_available": tools,
            "tools_called": mot.tool_calls,
            "seed": seed,
        }
        generate_log.action = mot._action
        generate_log.result = mot

        mot._generate_log = generate_log

    @overload
    async def generate_from_raw(
        self,
        actions: list[Component[C]],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk[C]]: ...

    @overload
    async def generate_from_raw(
        self,
        actions: list[Component[C] | CBlock],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk[C | str]]: ...

    async def generate_from_raw(
        self,
        actions: Sequence[Component[C] | CBlock],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk]:
        """Generate completions for multiple actions without chat templating.

        Passes formatted prompt strings directly to the HuggingFace model's
        ``generate`` method as a batch. Tool calling is not supported.

        Args:
            actions (Sequence[Component[C] | CBlock]): Actions to generate completions for.
            ctx (Context): The current generation context.
            format (type[BaseModelSubclass] | None): Optional Pydantic model for
                structured output decoding via llguidance.
            model_options (dict | None): Per-call model options.
            tool_calls (bool): Ignored; tool calling is not supported on this endpoint.

        Returns:
            list[ModelOutputThunk]: A list of model output thunks, one per action.
        """
        with instrument_generate_from_raw(
            backend=self, num_actions=len(actions), format=format, tool_calls=tool_calls
        ):
            await self.do_generate_walks(list(actions))

            if tool_calls:
                FancyLogger.get_logger().warning(
                    "The raw endpoint does not support tool calling at the moment."
                )

            if self._model.device.type == "mps":
                # TODO: Remove this when we are able to update the torch package.
                #       Test this by ensuring all outputs from this call are populated when running on mps.
                #       https://github.com/pytorch/pytorch/pull/157727
                FancyLogger.get_logger().warning(
                    "utilizing device mps with a `generate_from_raw` request; you may see issues when submitting batches of prompts to a huggingface backend; ensure all ModelOutputThunks have non-empty values."
                )

            model_opts = self._simplify_and_merge(model_options)
            seed = model_opts.get(ModelOption.SEED, None)
            if seed is not None:
                set_seed(seed)

            prompts = [self.formatter.print(action) for action in actions]

            # batch-encoding call is deprecated in favor of this
            inputs = self._tokenizer(prompts, return_tensors="pt", padding=True).to(
                self._device
            )

            format_kwargs = {}
            if format:
                schema: dict[str, Any] = format.model_json_schema()
                grammar: str = llguidance.LLMatcher.grammar_from_json_schema(
                    schema, defaults={"whitespace_flexible": False}
                )
                logits_processor = _GuidanceLogitsProcessor(
                    grammar, self._llguidance_tokenizer
                )
                format_kwargs["logits_processor"] = LogitsProcessorList(
                    [logits_processor]
                )

            outputs = await asyncio.to_thread(
                self._generate_with_adapter_lock,
                "",  # Empty for no adapter.
                self._model.generate,  # type: ignore
                # Passed as args/kwargs to generate.
                input_ids=inputs["input_ids"],
                attention_mask=inputs["attention_mask"],
                return_dict_in_generate=True,
                output_scores=True,
                **self._make_backend_specific_and_remove(model_opts),
                **format_kwargs,
            )

        sequences_to_decode = [
            sequence[inputs["input_ids"][i].size(0) :]  # type: ignore
            for i, sequence in enumerate(outputs.sequences)
        ]

        decoded_results = self._tokenizer.batch_decode(
            sequences_to_decode, skip_special_tokens=True
        )

        results = []
        for i, decoded_result in enumerate(decoded_results):
            n_prompt_tokens = inputs["input_ids"][i].size(0)  # type: ignore
            n_completion_tokens = len(sequences_to_decode[i])
            result = ModelOutputThunk(
                value=decoded_result,
                meta={
                    "usage": {
                        "prompt_tokens": n_prompt_tokens,  # type: ignore
                        "completion_tokens": n_completion_tokens,
                        "total_tokens": n_prompt_tokens + n_completion_tokens,
                    }
                },
            )
            action = actions[i]
            result.parsed_repr = (
                action.parse(result) if isinstance(action, Component) else result.value
            )

            generate_log = GenerateLog()
            generate_log.prompt = self.formatter.print(actions[i])
            generate_log.backend = f"hf::{self.model_id!s}"
            generate_log.model_options = model_opts
            generate_log.date = datetime.datetime.now()
            generate_log.model_output = decoded_result
            generate_log.extra = {"format": format, "seed": seed}
            generate_log.action = action

            result._generate_log = generate_log
            results.append(result)

        return results

    # region cache management
    def cache_get(self, id: str | int) -> HFAloraCacheInfo | None:
        """Retrieve a cached ``HFAloraCacheInfo`` entry by its key.

        Args:
            id (str | int): The cache key to look up.

        Returns:
            HFAloraCacheInfo | None: The cached entry, or ``None`` if not found.
        """
        v = self._cache.get(id)
        assert v is None or type(v) is HFAloraCacheInfo
        return v

    def cache_put(self, id: str | int, v: HFAloraCacheInfo):
        """Store an ``HFAloraCacheInfo`` entry in the cache under the given key.

        Args:
            id (str | int): The cache key to store the entry under.
            v (HFAloraCacheInfo): The cache entry containing KV cache state
                and associated generation metadata.
        """
        self._cache.put(id, v)

    # endregion

    def _simplify_and_merge(
        self, model_options: dict[str, Any] | None
    ) -> dict[str, Any]:
        """Simplifies model_options to use the Mellea specific ModelOption.Option and merges the backend's model_options with those passed into this call.

        Rules:
        - Within a model_options dict, existing keys take precedence. This means remapping to mellea specific keys will maintain the value of the mellea specific key if one already exists.
        - When merging, the keys/values from the dictionary passed into this function take precedence.

        Because this function simplifies and then merges, non-Mellea keys from the passed in model_options will replace
        Mellea specific keys from the backend's model_options.

        Common model options: https://huggingface.co/docs/transformers/en/llm_tutorial#common-options

        Args:
            model_options: the model_options for this call

        Returns:
            a new dict
        """
        backend_model_opts = ModelOption.replace_keys(
            self.model_options, self.to_mellea_model_opts_map
        )

        if model_options is None:
            return backend_model_opts

        generate_call_model_opts = ModelOption.replace_keys(
            model_options, self.to_mellea_model_opts_map
        )
        return ModelOption.merge_model_options(
            backend_model_opts, generate_call_model_opts
        )

    def _make_backend_specific_and_remove(
        self, model_options: dict[str, Any]
    ) -> dict[str, Any]:
        """Maps specified Mellea specific keys to their backend specific version and removes any remaining Mellea keys.

        Args:
            model_options: the model_options for this call

        Returns:
            a new dict
        """
        backend_specific = ModelOption.replace_keys(
            model_options, self.from_mellea_model_opts_map
        )
        return ModelOption.remove_special_keys(backend_specific)

    def _filter_chat_template_only_options(
        self, model_options: dict[str, Any]
    ) -> dict[str, Any]:
        """Remove options that are only for apply_chat_template, not for generate().

        Args:
            model_options: the model_options for this call

        Returns:
            a new dict without chat template-specific options
        """
        # Options that should only go to apply_chat_template, not generate()
        chat_template_only = {
            "guardian_config",
            "think",
            "add_generation_prompt",
            "documents",
        }
        return {k: v for k, v in model_options.items() if k not in chat_template_only}

    # region Adapter loading, unloading, and utility functions.
    @property
    def base_model_name(self):
        """Returns the base_model_id of the model used by the backend. For example, `granite-3.3-8b-instruct` for `ibm-granite/granite-3.3-8b-instruct`."""
        return self._hf_model_id.split("/")[1]

    def add_adapter(self, adapter: LocalHFAdapter):
        """Register a LoRA/aLoRA adapter with this backend so it can be loaded later.

        Downloads the adapter weights (via ``adapter.get_local_hf_path``) and records
        the adapter in the backend's registry. The adapter must not already be
        registered with a different backend.

        Args:
            adapter (LocalHFAdapter): The adapter to register with this backend.

        Raises:
            Exception: If ``adapter`` has already been added to a different backend.
        """
        if adapter.backend is not None:
            if adapter.backend is self:
                FancyLogger.get_logger().warning(
                    f"attempted to add adapter {adapter.name} with type {adapter.adapter_type} to the same backend {adapter.backend}"
                )
                return
            else:
                raise Exception(
                    f"adapter {adapter.name} with type {adapter.adapter_type} has already been added to backend {adapter.backend}"
                )

        if self._added_adapters.get(adapter.qualified_name) is not None:
            FancyLogger.get_logger().warning(
                f"Client code attempted to add {adapter.name} with type {adapter.adapter_type} but {adapter.name} was already added to {self.__class__}. The backend is refusing to do this, because adapter loading is not idempotent."
            )
            return None

        adapter.path = adapter.get_local_hf_path(self.base_model_name)
        adapter.backend = self
        self._added_adapters[adapter.qualified_name] = adapter

    def load_adapter(self, adapter_qualified_name: str):
        """Load a previously registered adapter into the underlying HuggingFace model.

        The adapter must have been registered via ``add_adapter`` first. Do not call
        this method while generation requests are in progress.

        Args:
            adapter_qualified_name (str): The ``adapter.qualified_name`` of the adapter
                to load (i.e. ``"<name>_<adapter_type>"``)

        Raises:
            ValueError: If no adapter with the given qualified name has been added to
                this backend.
        """
        adapter = self._added_adapters.get(adapter_qualified_name, None)
        if adapter is None:
            raise ValueError(
                f"could not load adapter {adapter_qualified_name} for backend {self}: adapter was not previously added"
            )

        try:
            adapter_kwargs = {}

            # Peft tries to stringify the device. If it's mps, it gets stringified as "mps:0" which causes
            # an error when loading with safetensors.torch.load_file. Force the device as a string "mps" to fix.
            if self._device == torch.device("mps"):
                adapter_kwargs["device"] = "mps"
            self._model.load_adapter(
                adapter.path, adapter.qualified_name, adapter_kwargs=adapter_kwargs
            )
        except ValueError as e:
            # If it's just that it's already loaded, ignore it.
            if f"Adapter with name {adapter_qualified_name} already exists." not in str(
                e
            ):
                raise e

        # Loading an adapter activates it. We disable adapters immediately after.
        # Prefer this over `.disable_adapters()`; the disable function doesn't always
        # seem to work.
        self._model.set_adapter([])
        # self._model.disable_adapters()
        self._loaded_adapters[adapter.qualified_name] = adapter

    def unload_adapter(self, adapter_qualified_name: str):
        """Unload a previously loaded adapter from the underlying HuggingFace model.

        If the adapter is not currently loaded, a log message is emitted and the
        method returns without error.

        Args:
            adapter_qualified_name (str): The ``adapter.qualified_name`` of the adapter
                to unload.
        """
        # Check if the backend knows about this adapter.
        adapter = self._loaded_adapters.get(adapter_qualified_name, None)
        if adapter is None:
            FancyLogger.get_logger().info(
                f"could not unload adapter {adapter_qualified_name} for backend {self}: adapter is not loaded"
            )
            return

        self._model.delete_adapter(adapter.qualified_name)

        # Remove the adapter from the list of loaded adapters.
        del self._loaded_adapters[adapter.qualified_name]

    def list_adapters(self) -> list[str]:
        """List the qualified names of all adapters currently loaded in this backend.

        Returns:
            list[str]: Qualified adapter names (i.e. ``adapter.qualified_name``) for
                all adapters that have been loaded via ``load_adapter``.
        """
        return list(self._loaded_adapters.keys())


def _assert_correct_adapters(expected_state: str, model: PreTrainedModel):
    """When generating with a huggingface model, this can be used to ensure the correct adapters are active.

    Args:
        expected_state: the current state of the lock
        model: the model underlying the LocalHFBackend; this is the model the adapters are activated on
    """
    try:
        active = model.active_adapters()

        if expected_state == "":
            assert len(active) == 0, (
                f'no adapters should be active if expected state is "", got "{active[0]}"'
            )
        else:
            assert len(active) == 1, (
                f'one adapter should be active if expected state is "{expected_state}"'
            )
            assert active[0] == expected_state, (
                f'the active adapter "{active[0]}" doesn\'t match the expected state: "{expected_state}"'
            )
    except ValueError as e:
        # If no weights have been loaded, the model will raise a ValueError:
        # `ValueError("No adapter loaded. Please load an adapter first.")`
        if "No adapter loaded" in str(e):
            assert expected_state == "", (
                f'got no adapters loaded but expected state is "{expected_state}"'
            )
        else:
            raise e



from collections.abc import Sequence
from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from ._icl_examples import ICLExample, icl_examples as default_icl_examples

this_file_dir = Path(__file__).resolve().parent

environment = Environment(loader=FileSystemLoader(this_file_dir), autoescape=False)
system_template = environment.get_template("system_template.jinja2")
user_template = environment.get_template("user_template.jinja2")


def get_system_prompt(icl_examples: list[ICLExample] = default_icl_examples) -> str:
    return system_template.render(icl_examples=icl_examples).strip()


def get_user_prompt(
    execution_plan: list[str],
    constraint_list: Sequence[str],
    subtask_title: str,
    subtask_prompt: str,
) -> str:
    return user_template.render(
        execution_plan=execution_plan,
        constraint_list=constraint_list,
        subtask_title=subtask_title,
        subtask_prompt=subtask_prompt,
    ).strip()



"""``TemplateFormatter``: Jinja2-template-based formatter for legacy backends.

``TemplateFormatter`` extends ``ChatFormatter`` to look up a per-component Jinja2
template at rendering time, allowing each ``Component`` type to control its own
prompt representation. Template discovery walks a configurable ``template_path`` and
the built-in templates directory; results are cached in a ``SimpleLRUCache`` for
performance. Use this formatter when your backend requires hand-crafted prompts
rather than a generic chat-message rendering.
"""

import os
import re
import sys
from collections.abc import Iterable, Mapping
from dataclasses import fields
from typing import Any

import jinja2

from ..backends.cache import SimpleLRUCache
from ..backends.model_ids import ModelIdentifier
from ..core import CBlock, Component, FancyLogger, TemplateRepresentation
from .chat_formatter import ChatFormatter


class TemplateFormatter(ChatFormatter):
    """Formatter that uses Jinja2 templates to render components into prompt strings.

    Template discovery walks a configurable ``template_path`` and the built-in
    templates directory. Results are optionally cached in a ``SimpleLRUCache``
    for performance. Use this formatter when your backend requires hand-crafted
    prompts rather than generic chat-message rendering.

    Args:
        model_id (str | ModelIdentifier): Describes the model for which templates will be looked up.
            Should match the template directory structure.
        template_path (str): An alternate location where templates can be found.
            Will be preferred over all other template directories even if a less exact match is found.
            Defaults to ``""``.
        use_template_cache (bool): When ``True``, caches template lookup results. Set to ``False``
            if you plan to change ``model_id`` or ``template_path`` after construction.
            Defaults to ``True``.

    Example::

        formatter = TemplateFormatter(model_id="my-model", template_path="/path/to/templates")
        text = formatter.print(my_component)
    """

    def __init__(
        self,
        model_id: str | ModelIdentifier,
        *,
        template_path: str = "",
        use_template_cache: bool = True,
    ):
        """Initialize TemplateFormatter for the given model ID with optional template path and caching."""
        self.model_id = model_id
        self._template_path: str = template_path

        self._use_template_cache: bool = use_template_cache

        # Key: obj.__class__.__name___ -> Value: jinja2.Template
        self._template_cache = SimpleLRUCache(10) if self._use_template_cache else None

    def _stringify(
        self,
        c: (
            str
            | Component
            | CBlock
            | Iterable
            | Mapping
            | TemplateRepresentation
            | None
        ),
    ) -> Any:
        """A recursive function that ensures an object is stringified.

        For strings and CBlocks, this is just getting their value.
        For Components, this means traversing the fields of their template args dictionary and stringifying each part.

        Iterables and Mappings should only be encountered as parts of a component's template args. We process each item in them while maintaining the structure.
        """
        if c is None:
            return ""

        match c:
            case str():
                return c

            case CBlock():
                assert c.value is not None
                return c.value

            case Component():
                representation = c.format_for_llm()
                if type(representation) is str:
                    return representation
                else:
                    assert isinstance(representation, TemplateRepresentation)
                    stringified_template_args = {}
                    for key, val in representation.args.items():
                        stringified_template_args[key] = self._stringify(val)

                    if representation.obj is None:
                        FancyLogger.get_logger().warning(
                            f"template formatter encountered a TemplateRepresentation with no obj when stringifying {c.__class__}; setting obj to {c}"
                        )
                        representation.obj = c
                    return self._load_template(representation).render(
                        stringified_template_args
                    )

            case c if isinstance(c, Mapping):
                stringified_template_args = {}
                for key, val in c.items():
                    stringified_template_args[key] = self._stringify(val)

                return stringified_template_args

            case c if isinstance(c, Iterable) and not isinstance(c, Mapping):
                stringified_list = []
                for val in c:
                    stringified_list.append(self._stringify(val))

                return stringified_list

            case _:
                FancyLogger.get_logger().warning(
                    f"formatter encountered an unexpected type in _stringify; using str() on {c.__class__}"
                )
                return str(c)

    def print(self, c: Component | CBlock) -> str:
        """Render a component or code block to a string using a Jinja2 template.

        Args:
            c (Component | CBlock): The component or code block to render.

        Returns:
            str: The rendered string representation of the component.
        """
        return self._stringify(c)

    def _load_template(self, repr: TemplateRepresentation) -> jinja2.Template:
        """This method makes an attempt at auto-loading a Template for the Component.

        Iterates over template order in the template representation. If it finds
        a '*' it will check for a template with the name of the object in 'obj'.
        Once it finds a template, it stops searching the list.

        To find a template, it looks for it in the following places in order:
        1. the user provided template location
        2. the component's package or the mellea package if none is found

        Raises:
            Exception: If there's an unexpected jinja error or the template cannot be found.
        """
        if repr.template:
            return jinja2.Environment().from_string(repr.template)  # type: ignore

        if repr.template_order is None:
            FancyLogger.get_logger().warning(
                f"using template formatter for {repr.obj.__class__.__name__} but no template order was provided, defaulting to class name"
            )
            repr.template_order = ["*"]

        # If using a cache, check the cache for the loaded template first.
        if self._use_template_cache and self._template_cache:
            res = self._template_cache.get(repr.obj.__class__.__name__)
            if res:
                return res

        qualified_tmpl_name = (
            ""  # Includes some basic path info: `prompts/default/Instruction.jinja2`.
        )
        loader: jinja2.PackageLoader | jinja2.FileSystemLoader | None = None

        # Iterate over all possible template names to find a match.
        for t_name in repr.template_order:
            if t_name == "*":
                tmpl_name = repr.obj.__class__.__name__ + ".jinja2"
            else:
                tmpl_name = t_name + ".jinja2"

            if self._template_path != "":
                # Look for the template at the user specified path if it exists.
                loader = jinja2.FileSystemLoader(self._template_path)
                qualified_tmpl_name = self._get_template(self._template_path, tmpl_name)

            if qualified_tmpl_name == "":
                # Try to get the package name to load the templates from it.
                # Default to Mellea if one can't be found.
                package = _get_package_name(repr.obj.__class__.__module__)
                try:
                    loader = jinja2.PackageLoader(package)
                except Exception as e:
                    if package != "mellea":
                        # Mellea should always be available.
                        loader = jinja2.PackageLoader("mellea")
                    else:
                        raise ValueError(
                            f"could not find package for obj: {repr.obj}, exception: {e}"
                        )

                qualified_tmpl_name = self._get_template(
                    loader._template_root, tmpl_name
                )

            # If the qualitified template name isn't an empty string,
            # we've found a template and its loader.
            if qualified_tmpl_name != "":
                break

        if qualified_tmpl_name == "":
            raise ValueError(
                f"could not find template candidate for {repr.obj}, searched: {repr.template_order}"
            )

        try:
            # If we get here, we know the loader is populated and we have found a file that
            # contains the template for the object.
            env = jinja2.Environment(
                loader=loader, autoescape=jinja2.select_autoescape()
            )
            tmpl = env.get_template(qualified_tmpl_name)

        except (jinja2.TemplateNotFound, ValueError) as tnf:
            # This shouldn't happen since we specifically walk the directory tree to get the best template, but
            # there's a chance some file system issue happens.
            raise Exception(
                f"Could not find a template for {repr.obj} using template {qualified_tmpl_name}: {tnf}"
            )

        except Exception as e:
            raise Exception(f"Unexpected template error for component {repr.obj}: {e}")

        if self._use_template_cache and self._template_cache:
            self._template_cache.put(repr.obj.__class__.__name__, tmpl)
        return tmpl

    def _get_template(self, root_path: str, template_name: str) -> str:
        """Attempts to walk the provided directory structure to find the best matching template.

        Prefers the most exact match (meaning deepest directory tree) by:
        1. Looking for directory names matching the model name
        2. Looking in the `prompts/default/` directory for a template

        Assumes that only one directory at each level matches.
        """
        # Simplify the provided model to make matching easier.
        model_id = self._get_model_id()
        simplified_model_id = _simplify_model_string(model_id)

        # Traverse the directory structure to find the most specific template.
        path_offset = len(root_path)  # Used to get the template name used by Jinja2.
        candidate_template = ""
        for root, dirs, files in os.walk(top=root_path, topdown=True):
            # Only look at the default templates if a candidate hasn't yet been chosen.
            # Otherwise, we already have a more specific template.
            is_default = root.rsplit("/")[-1].lower() == "default"
            candidate_found = candidate_template != ""
            if not (is_default and candidate_found):
                for file in files:
                    if file.lower() == template_name.lower():
                        template_path = root[path_offset:].lstrip(os.path.sep)
                        candidate_template = os.path.join(template_path, file).replace(
                            os.path.sep, "/"
                        )

            # Only traverse file paths that are in the model id or are prompts/default.
            for dir in dirs:
                if dir.lower() == "prompts" or dir.lower() == "default":
                    continue

                # Simplify the directory as well for matching names.
                if _simplify_model_string(dir.lower()) not in simplified_model_id:
                    dirs.remove(dir)

        return candidate_template

    def _get_model_id(self) -> str:
        """Gets a string representation of the formatter's model id."""
        model_id = self.model_id
        if type(model_id) is str:
            return model_id

        assert isinstance(model_id, ModelIdentifier), (
            "model_id was neither a `str` nor `ModelIdentifier`"
        )

        # Go through the ModelIdentifier's fields, find one that can be matched against.
        for field in fields(model_id):
            val = getattr(model_id, field.name)
            if val is not None and val != "":
                return val

        return ""  # Cannot match against any model identifiers. Will ultimately use default.


def _simplify_model_string(input: str) -> str:
    """Removes special chars from the given string and lower cases it to simplify matching."""
    remove_chars = r"[-\.\/:,]"
    return re.sub(remove_chars, "", input.lower())


def _get_package_name(module: str) -> str:
    """Given a module, attempts to get the package and verifies it exists."""
    package = module.split(".")[0]

    if sys.modules.get(package, None) is None:
        package = ""

    return package
